#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command-line interface for downloading records."""

from asyncio import Task, TaskGroup
from pathlib import Path
from typing import Any, Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client import limit_connections
from nrp_cmd.async_client.streams import FileSink
from nrp_cmd.cli.base import OutputFormat, async_command
from nrp_cmd.cli.records.get import get_single_record
from nrp_cmd.config import Config
from nrp_cmd.progress import show_progress

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    with_config,
    with_model,
    with_output,
    with_progress,
    with_record_ids,
    with_repository,
    with_verbosity,
)


@click.option("--expand", is_flag=True, help="Expand the record")
@with_config
@with_repository
@with_record_ids
@with_verbosity
@with_output
@with_progress
@with_model
@async_command
async def download_record(
    *,
    config: Config,
    repository: Optional[str],
    record_ids: list[str],
    out: Output,
    model: Model,
    expand: bool = False,
) -> None:
    """Download a record from the repository.

    The metadata of the record are stored together with the files in output directory.
    """
    console = Console()
    tasks: list[Task[Any]] = []
    with (
        limit_connections(10),
        show_progress(total=len(record_ids), quiet=not out.progress),
    ):
        async with TaskGroup() as tg:
            for record_id in record_ids:
                tasks.append(
                    tg.create_task(
                        download_single_record(
                            record_id,
                            console,
                            config,
                            repository,
                            model.model,
                            out.output,
                            out.output_format,
                            model.published,
                            model.draft,
                            expand,
                            out.verbosity,
                        )
                    )
                )
    abort = False
    for task in tasks:
        if task.result() is False:
            abort = True

    if abort:
        console.print("[red]Failed to download some of the records[/red]")
        raise click.Abort()


async def download_single_record(
    record_id: str,
    console: Console,
    config: Config,
    repository: str | None,
    model: str | None,
    output: Path | None,
    output_format: OutputFormat | None,
    published: bool,
    draft: bool,
    expand: bool,
    verbosity: VerboseLevel,
) -> bool:
    """Download record with the given id together with its files."""
    # 1. download record metadata
    if not output:
        output = Path("{id}")

    if not output_format:
        output_format = OutputFormat.JSON

    record, output, _repository_config, repository_client = await get_single_record(
        record_id,
        console,
        config,
        repository,
        model,
        output / "metadata{ext}",
        output_format,
        published,
        draft,
        expand,
        verbosity=VerboseLevel.NORMAL,
    )

    if not record:
        return False

    output_dir = output.parent if output and output.parent else Path.cwd()

    assert repository_client
    file_client = repository_client.files

    file_list = await file_client.list(record)

    tasks: list[Any] = []
    async with TaskGroup() as tg:
        for file_ in file_list:
            file_key = file_.key
            # sanitize the key
            if "/" in file_key:
                file_key = file_key.replace("/", "_")
            file_key = file_key.replace(":", "_")
            file_key = file_key.replace("..", "_")
            tasks.append(
                tg.create_task(
                    file_client.download(
                        file_, FileSink(output_dir / file_key), progress=file_.key
                    )
                )
            )

    ok = True
    for idx, task in enumerate(tasks):
        if task.exception():
            ok = False
            console.print(
                f"[red]Failed to download file {file_list[idx].key} of {record_id}[/red]"
            )

    if ok and verbosity != VerboseLevel.QUIET:
        console.print(f"[green]Record {record_id} downloaded to {output_dir}[/green]")
    return ok



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Variables mainly for storing record urls."""

import json
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

from attrs import define, field

from ..converter import converter


@define(kw_only=True)
class Variables:
    """Variables for the commandline tools."""

    variables: dict[str, list[str]] = field(factory=dict)
    """Internal dictionary of variables."""

    _config_file_path: Optional[Path] = None
    """Path to the configuration file from which the variables have been loaded."""

    class Config:  # noqa
        extra = "forbid"

    @classmethod
    def from_file(cls, config_file_path: Optional[Path] = None) -> "Variables":
        """Load the configuration from a file."""
        if not config_file_path:
            config_file_path = Path.home() / ".nrp" / "variables.json"

        if config_file_path.exists():
            ret = converter.structure(
                json.loads(config_file_path.read_text(encoding="utf-8")),
                cls
            )
        else:
            ret = cls()
        ret._config_file_path = config_file_path
        return ret

    def save(self, path: Optional[Path] = None) -> None:
        """Save the configuration to a file, creating parent directory if needed."""
        if path:
            self._config_file_path = path
        else:
            path = self._config_file_path
        assert path, "No path to save the configuration to."
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(
            json.dumps(converter.unstructure(self), indent=2, ensure_ascii=False),
            encoding="utf-8",
        )

    def __getitem__(self, key: str) -> list[str]:
        """Get the value of a variable, raising KeyError if the variable has not been found."""
        try:
            return self.variables[key]
        except KeyError as err:
            raise KeyError(
                f"Variable {key} not found at {self._config_file_path}"
            ) from err

    def __setitem__(self, key: str, value: list[str]):
        """Set the value of a variable."""
        self.variables[key] = value

    def __delitem__(self, key: str):
        """Delete a variable."""
        del self.variables[key]

    def get(self, key: str, default: list[str] | None = None) -> list[str] | None:
        """Get the value of a variable, returning None if the variable has not been found."""
        return self.variables.get(key, default)

    def items(self) -> Iterable[tuple[str, list[str]]]:
        """Return the variables as an iterable of (key, value)."""
        return self.variables.items()



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline clients for working with variables."""


from typing import Any

from rich import box
from rich.console import Console
from rich.table import Table

from nrp_cmd.config import Config

from .arguments import Output, argument_with_help, with_config, with_output
from .base import OutputWriter


@with_config
@argument_with_help("name", type=str, help="The name of the variable")
@argument_with_help("values", type=str, nargs=-1, help="The values of the variable")
def set_variable(*, config: Config, name: str, values: list[str]) -> None:
    """Add a variable to the configuration."""
    console = Console()
    console.print()
    variables = config.load_variables()

    variables[name] = values
    variables.save()
    msg = f"Added variable {name} -> "
    ln = len(msg)
    formatted_values = (f"\n{' ' * ln}".join(values)).strip()
    console.print(f"[green]{msg}{formatted_values}[/green]")


@with_config
@argument_with_help("name", type=str, help="The name of the variable")
def remove_variable(config: Config, name: str) -> None:
    """Remove a variable from the configuration."""
    console = Console()
    console.print()
    variables = config.load_variables()

    del variables[name]
    variables.save()

    console.print(f"[green]Removed variable {name}[/green]")


@with_config
@with_output
def list_variables(*, config: Config, out: Output) -> None:
    """List all variables."""
    console = Console()
    variables = config.load_variables()

    with OutputWriter(
        out.output, out.output_format, console, variables_table
    ) as printer:
        printer.output(variables.variables)


@with_config
@with_output
@argument_with_help("variable", type=str, help="The name of the variable")
def get_variable(*, config: Config, variable: str, out: Output) -> None:
    """Get all variables."""
    console = Console()
    variables = config.load_variables()
    value = variables.get(variable)

    def variable_table(data: list[str], **kwargs: Any) -> str:
        return "\n".join(data)

    with OutputWriter(
        out.output, out.output_format, console, variable_table
    ) as printer:
        printer.output(value)


def variables_table(data: dict[str, list[str]], **kwargs: Any) -> Table:
    """Render a table of variables."""
    table = Table(title="Variables", box=box.SIMPLE, title_justify="left")
    table.add_column("Name", style="cyan")
    table.add_column("Values")
    for name, values in data.items():
        table.add_row(name, "\n".join(values))
    return table



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Local transfer."""

from __future__ import annotations

from asyncio import TaskGroup
from typing import TYPE_CHECKING

from yarl import URL

from ...connection.aws_limits import adjust_upload_multipart_params
from ...streams.progress import ProgressSource
from . import Transfer

if TYPE_CHECKING:
    from yarl import URL

    from ....progress import ProgressBar
    from ....types.files import File, MultipartUploadLinks
    from ...connection import AsyncConnection
    from ...streams import DataSource


class MultipartTransfer(Transfer):
    """Local transfer.

    This transfer copies a local file to the repository.
    The file will be stored in repository's primary storage (thus local)
    and the upload will be handled solely through the repository.
    """

    async def prepare(
        self,
        connection: AsyncConnection,
        files_link: URL,
        transfer_payload: dict,
        source: DataSource,
    ) -> None:
        """Prepare the transfer."""
        if not transfer_payload.get("size"):
            transfer_payload["size"] = await source.size()

        transfer_md = transfer_payload.get("transfer", {})

        parts, part_size = adjust_upload_multipart_params(
            transfer_payload["size"],
            transfer_md.get("parts"),
            transfer_md.get("part_size"),
        )
        transfer_md["parts"] = parts
        transfer_md["part_size"] = part_size
        
    async def upload(
        self,
        connection: AsyncConnection,
        initialized_upload: File,
        source: DataSource,
        progress_bar: ProgressBar,
    ) -> None:
        """Upload the file."""
        links: list[MultipartUploadLinks] = initialized_upload.links.parts or []
        assert links

        number_of_parts = len(links)
        part_size: int = initialized_upload.transfer.part_size

        size = initialized_upload.size
        progress_bar.set_total(size)

        async def put_stream(pt: int, start: int, count: int):
            headers: dict[str, str] = {
                "Content-Length": str(count),
                "Content-Type": "application/octet-stream",
            }
            if "md5" in source.supported_checksums():
                headers["Content-MD5"] = await source.checksum(
                    "md5", offset=start, count=count
                )
            return await connection.put_stream(
                url=links[pt].url,
                source=ProgressSource(source, progress_bar),
                open_kwargs={"offset": start, "count": count},
                headers=headers,
            )

        async with TaskGroup() as tg:
            for pt in range(number_of_parts):
                start = pt * part_size
                count = min(part_size, size - start)
                tg.create_task(put_stream(pt, start, count))

    async def get_commit_payload(self, initialized_upload: File) -> dict:
        """Get payload for finalization of the successful upload."""
        return {}



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line interface for getting records."""

import asyncio
from functools import partial
from pathlib import Path
from typing import Optional

import rich_click as click
from rich.console import Console
from yarl import URL

from nrp_cmd.async_client import (
    AsyncRecordsClient,
    AsyncRepositoryClient,
    get_async_client,
    get_repository_from_record_id,
)
from nrp_cmd.async_client.connection import AsyncConnection, limit_connections
from nrp_cmd.cli.base import OutputFormat, OutputWriter, async_command
from nrp_cmd.cli.records.record_file_name import create_output_file_name
from nrp_cmd.cli.records.table_formatters import format_record_table
from nrp_cmd.config import Config, RepositoryConfig
from nrp_cmd.errors import DoesNotExistError
from nrp_cmd.types.records import Record

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    with_config,
    with_model,
    with_output,
    with_record_ids,
    with_repository,
    with_verbosity,
)


@with_config
@with_repository
@with_record_ids
@with_output
@with_verbosity
@with_model
@click.option("--expand", is_flag=True, help="Expand the record")
@async_command
async def get_record(
    config: Config,
    repository: Optional[str],
    record_ids: list[str],
    out: Output,
    model: Model,
    expand: bool = False,
) -> None:
    """Get a record from the repository."""
    console = Console()

    with limit_connections(10):
        tasks: list[
            asyncio.Task[
                tuple[
                    Record | None,
                    Path | None,
                    RepositoryConfig | None,
                    AsyncRepositoryClient | None,
                ]
            ]
        ] = []
        async with asyncio.TaskGroup() as tg:
            for record_id in record_ids:
                tasks.append(
                    tg.create_task(
                        get_single_record(
                            record_id,
                            console,
                            config,
                            repository,
                            model.model,
                            out.output,
                            out.output_format,
                            model.published,
                            model.draft,
                            expand,
                            out.verbosity,
                        )
                    )
                )
        results = [x.result() for x in tasks]
        for r in results:
            if r[0] is None:
                raise click.Abort()


async def get_single_record(
    record_id: str,
    console: Console,
    config: Config,
    repository: str | None,
    model: str | None,
    output: Path | None,
    output_format: OutputFormat | None,
    published: bool,
    draft: bool,
    expand: bool,
    verbosity: VerboseLevel,
) -> tuple[
    Record | None, Path | None, RepositoryConfig | None, AsyncRepositoryClient | None
]:
    """Get a single record from the repository and print/save it."""
    try:
        (
            record,
            _final_record_id,
            repository_config,
            _record_client,
            repository_client,
        ) = await read_record(
            record_id, repository, config, expand, model, published, draft
        )
    except DoesNotExistError as e:
        console.print(f"[red]Record with id {record_id} does not exist.[/red]")
        if verbosity == VerboseLevel.VERBOSE:
            print(e)
        return None, output, None, None

    if output:
        output = create_output_file_name(
            output, str(record.id or record_id or "unknown_id"), record, output_format
        )
    if output and output.parent:
        output.parent.mkdir(parents=True, exist_ok=True)

    # note: this is synchronous, but it is not a problem as only metadata are printed/saved
    with OutputWriter(
        output,
        output_format,
        console,
        partial(format_record_table, verbosity=verbosity),  # type: ignore # mypy does not understand this
    ) as printer:
        printer.output(record)

    return record, output, repository_config, repository_client


async def read_record(
    record_id: str,
    repository: str | None,
    config: Config,
    expand: bool,
    model: str | None,
    published: bool,
    draft: bool,
) -> tuple[
    Record, str | URL, RepositoryConfig, AsyncRecordsClient, AsyncRepositoryClient
]:
    """Read a record from the repository, returning the record, its id and the repository config."""
    connection = AsyncConnection()

    final_record_id, repository_config = await get_repository_from_record_id(
        connection, record_id, config, repository
    )
    # set it temporarily to the config
    config.add_repository(repository_config)
    client = await get_async_client(repository, config=config)
    records_api: AsyncRecordsClient = client.records
    if model is not None:
        records_api = records_api.with_model(model)
    if published:
        records_api = records_api.published_records
    if draft:
        records_api = records_api.draft_records
    query: dict[str, str] = {}
    if expand:
        query["expand"] = "true"

    record = await records_api.read(record_id=final_record_id, query=query)
    return record, final_record_id, repository_config, records_api, client



#
# This file was generated from the asynchronous client at invenio/client.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Invenio asynchronous client."""

from typing import Any, Self, override

from yarl import URL  # noqa: TCH002    as attrs need to have type info in runtime

from ...config import RepositoryConfig
from ...errors import (
    RepositoryClientError,
    RepositoryCommunicationError,
    StructureError,
    is_instance_of_exceptions,
)
from ...rdm_compat import make_rdm_info
from ...types.info import ModelInfo, RepositoryInfo
from ..base_client import SyncRepositoryClient
from ..connection import SyncConnection
from .files import SyncInvenioFilesClient
from .records import SyncInvenioRecordsClient
from .requests import SyncInvenioRequestsClient


class SyncInvenioRepositoryClient(SyncRepositoryClient):
    """An abstract client for NRP repositories.

    Usually, subclasses of this class are not instantiated directly in your code.
    To get an instance of a repository, use the high-level method `get_sync_client`:

    ```
    my_client = sync_client(config?, url=url, refresh=False/True)
    my_client = sync_client(config?, alias=alias, refresh=False/True)
    ```

    and then use the instance.
    """

    @override
    @classmethod
    def can_handle_repository(
        cls, url: URL | str, verify_tls: bool = True
    ) -> URL | None:
        connection = SyncConnection(verify_tls=verify_tls)
        if isinstance(url, str):
            url = URL(url)

        # check if the repository is an NRP invenio repository
        well_known_url = url.with_path("/.well-known/repository")
        try:
            data = connection.get(url=well_known_url, result_class=dict[str, Any])
            if "invenio_version" in data:
                if "api" in data.get("links", {}):
                    return URL(data["links"]["api"])
                # fallback for older oarepo-runtime versions
                return url.with_path("/api")
        except Exception:
            pass

        # check if the repository is a plain Invenio RDM repository, such as zenodo
        try:
            root_page_data = connection.get(url=url.with_path("/"), result_class=str)
            if '<meta name="generator" content="InvenioRDM' in root_page_data:
                return url.with_path("/api")
        except Exception:
            pass
        return None

    @override
    @classmethod
    def from_configuration(
        cls,
        config: RepositoryConfig,
        refresh: bool = False,
        extra_tokens: dict[URL, str] | None = None,
    ) -> Self:
        """Create a client from the given configuration.

        :param config: the configuration for the repository
        :param refresh: refresh the configuration by calling get_repository_info
        """
        ret = cls(config=config, extra_tokens=extra_tokens)
        if refresh or not config.info:
            config.info = ret.get_repository_info(refresh=True)
        return ret

    def __init__(
        self,
        config: RepositoryConfig,
        extra_tokens: dict[URL, str] | None = None,
    ):
        """Create a new client for the given repository configuration."""
        self._config = config
        tokens: dict[URL, str] = {}
        if extra_tokens:
            tokens.update(extra_tokens)
        if config.token:
            tokens[config.url] = config.token
        self._connection = SyncConnection(
            tokens=tokens,
            verify_tls=config.verify_tls,
            retry_count=config.retry_count,
            retry_after_seconds=config.retry_after_seconds,
        )

    @override
    def get_repository_info(self, refresh: bool = True) -> RepositoryInfo:
        if self._config.info and not refresh:
            return self._config.info

        try:
            info = self._connection.get(
                url=self._config.well_known_repository_url,
                result_class=RepositoryInfo,
            )
            self._config.info = info

            if not self._config.info.links.models:
                raise ValueError("The repository does not provide the models link.")

            models = self._connection.get(
                url=self._config.info.links.models,
                result_class=list[ModelInfo],
            )

            info.models = {
                model.type: model for model in models
            }

        except* (RepositoryClientError, RepositoryCommunicationError) as exc:
            if is_instance_of_exceptions(exc, StructureError):
                raise exc
            # not a NRP based repository, suppose that it is plain invenio rdm
            info = self._config.info = make_rdm_info(
                self._config.url, self._config.verify_tls
            )

        return info

    @property
    @override
    def records(self) -> SyncInvenioRecordsClient:
        """Return client for accessing records."""
        assert self._config.info, "Repository info is not available, can not create records client."
        return SyncInvenioRecordsClient(
            self._connection, self._config.info, self.requests
        )

    @property
    @override
    def files(self) -> SyncInvenioFilesClient:
        """Return client for accessing records."""
        assert self._config.info, "Repository info is not available, can not create records client."
        return SyncInvenioFilesClient(self._connection, self._config.info)

    @property
    @override
    def requests(self) -> SyncInvenioRequestsClient:
        """Return client for accessing requests."""
        assert (
            self._config.info
        ), "Repository info is not available, can not create requests client."
        return SyncInvenioRequestsClient(self._connection, self._config.info)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for files."""

from .delete import delete_file
from .download import download_files
from .list import list_files
from .update import update_file_metadata
from .upload import upload_files

__all__ = ("list_files", "download_files", "upload_files", "update_file_metadata",
           "delete_file")



#
# This file was generated from the asynchronous client at invenio/transfer/registry.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Registry of known transfer types."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ....types.files import TRANSFER_TYPE_LOCAL, TRANSFER_TYPE_MULTIPART

if TYPE_CHECKING:
    from .base import Transfer



class TransferRegistry:
    """Registry of known transfer types."""

    def __init__(self):
        """Initialize the registry."""
        self.transfers = {}

    def register(self, transfer_type: str, transfer: type[Transfer]) -> None:
        """Register a transfer type.

        :param transfer_type:       transfer type
        :param transfer:            registered transfer
        """
        self.transfers[transfer_type] = transfer

    def get(self, transfer_type: str) -> Transfer:
        """Get a transfer for a transfer type.

        :param transfer_type:       transfer type
        :return:                    instance of a transfer
        """
        return self.transfers[transfer_type]()


transfer_registry = TransferRegistry()
"""Singleton for the transfer registry"""

#
# supported transfers are registered here
#
from .local import LocalTransfer  # noqa
from .multipart import MultipartTransfer  # noqa

transfer_registry.register(TRANSFER_TYPE_LOCAL, LocalTransfer)
transfer_registry.register(TRANSFER_TYPE_MULTIPART, MultipartTransfer)




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line interface for accepting requests."""

from __future__ import annotations

from typing import Optional

from rich.console import Console

from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.config import Config

from ..arguments import (
    Output,
    argument_with_help,
    with_config,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)
from .table_formatter import format_request_table
from .utils import resolve_request


@with_config
@with_repository
@with_resolved_vars("request_id")
@argument_with_help("request_id", type=str, help="Request IDs")
@with_output
@with_verbosity
@async_command
async def accept_request(
    *, config: Config, repository: Optional[str] = None, request_id: str, out: Output
) -> None:
    """Accept a request."""
    console = Console()

    requests_client, request_url = await resolve_request(request_id, config, repository)

    request = await requests_client.accept(request_url)

    with OutputWriter(
        out.output, out.output_format, console, format_request_table
    ) as printer:
        printer.output(request)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Low-level compatibility layer for file operations."""

import asyncio
import base64
import hashlib
import os
from pathlib import Path
from typing import Literal, overload

import aiofiles
import aiofiles.os

from .base import InputStream, OutputStream


class FileInputStream(InputStream):
    async def seek(self, offset: int, whence: int = os.SEEK_SET) -> None:
        """Change the stream position."""
        ...
        
class FileOutputStream(OutputStream):
    async def seek(self, offset: int, whence: int = os.SEEK_SET) -> None:
        """Change the stream position."""
        ...

    async def truncate(self, size: int) -> None:
        ...
        
@overload
async def open_file(_fpath: Path, mode: Literal["rb"]) -> FileInputStream:
    ...

@overload
async def open_file(_fpath: Path, mode: Literal["wb"] | Literal["r+b"]) -> FileOutputStream:
    ...

async def open_file(_fpath: Path, mode: Literal["rb"] | Literal["wb"] | Literal["r+b"]) -> FileInputStream | FileOutputStream:
    """Open a file for reading or writing."""
    r: FileInputStream | FileOutputStream = await aiofiles.open(_fpath, mode=mode)  # noqa # type: ignore
    return r


async def file_stat(_fpath: Path) -> os.stat_result:
    """Get file statistics."""
    return await aiofiles.os.stat(_fpath)


async def checksum_file(
    file_name: Path, algo: str = "md5", offset: int = 0, count: int | None = None
) -> str:
    """Calculate the checksum of the file."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _checksum, file_name, algo, offset, count)


def _checksum(file_name: Path, algo: str, offset: int, count: int | None) -> str:
    """Calculate the checksum of the data."""
    # cpu-bound operation - run in a separate thread synchronously
    with open(file_name, mode="rb") as f:
        if offset > 0:
            f.seek(offset)

        hasher = hashlib.new(algo)
        chunk_size = get_filesystem_block_size(file_name) * 16

        while count != 0:
            chunk = f.read(min(chunk_size, count or chunk_size))
            if not chunk:
                break
            hasher.update(chunk)
            if count is not None:
                count -= len(chunk)

        return base64.b64encode(hasher.digest()).decode("ascii")


def get_filesystem_block_size(file_path: Path) -> int:
    if hasattr(os, "statvfs"):  # Unix-like systems
        statvfs = os.statvfs(str(file_path))
        return statvfs.f_bsize
    else:
        return 4096  # default block size for Windows


__all__ = (
    "open_file",
    "file_stat",
    "FileInputStream",
    "FileOutputStream",
    "checksum_file",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""A response that can raise parsed invenio errors."""

import json
from datetime import datetime
from email.utils import parsedate_to_datetime
from typing import Any

from aiohttp import ClientResponse

from ...errors import (
    DoesNotExistError,
    RepositoryClientError,
    RepositoryCommunicationError,
    RepositoryRetryError,
    RepositoryServerError,
)


class RepositoryResponse(ClientResponse):
    """A response that can raise parsed invenio errors."""

    async def raise_for_invenio_status(self) -> None:
        """Raise an exception if the response status code is not 2xx.

        :raises RepositoryServerError: if the status code is 5xx
        :raises RepositoryClientError: if the status code is 4xx
        :raises RepositoryCommunicationError: if the status code is not 2xx nor 4xx nor 5xx
        """
        if not self.ok:
            payload_text = await self.text()
            self.release()
            payload: Any
            try:
                payload = json.loads(payload_text)
            except ValueError:
                payload = {
                    "status": self.status,
                    "reason": payload_text,
                }
            if self.status == 429:
                after_seconds: float | None = 20
                retry_after = self.headers.get("Retry-After", None)
                if retry_after:
                    try:
                        after_seconds = float(retry_after)
                    except:
                        try:
                            after_datetime = parsedate_to_datetime(retry_after)
                            after_seconds = (
                                datetime.now() - after_datetime
                            ).total_seconds()
                        except:
                            pass
                raise RepositoryRetryError(after_seconds)

            if self.status >= 500:
                raise RepositoryServerError(self.request_info, payload)
            elif self.status >= 400:
                if self.status == 404:
                    raise DoesNotExistError(self.request_info, payload)
                raise RepositoryClientError(self.request_info, payload)
            raise RepositoryCommunicationError(self.request_info, payload)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""This package contains classes to work with the configuration of the NRP invenio client.

This configuration is stored in a JSON file
on a file system, usually in .nrp/invenio-config.json.

Normally you'd not use the classes directly, but instantiate the client directly
with get_sync_client / get_async_client functions from the nrp_cmd.client module.

To get a configuration object, you can use the Config.from_file() method
and either pass the path to the configuration file or let it default to ~/.nrp/invenio-config.json.
"""

from .config import Config
from .repository import RepositoryConfig

__all__ = ("Config", "RepositoryConfig")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Repository info endpoint response types."""

from typing import Optional

from attrs import define, field
from yarl import URL

from ..converter import Rename, extend_serialization
from . import Model


@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class RepositoryInfoLinks(Model):
    """Links within the repository info endpoint."""

    self_: URL = field()
    """Link to the repository itself"""

    records: URL = field()
    """Link to the global search endpoint"""

    drafts: URL | None = field(default=None)
    """Link to the user's records"""

    models: Optional[URL] = field(default=None)
    """Link to the models in the repository"""

    requests: URL | None = field(default=None)
    """Link to the requests in the repository"""
    

@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class ModelInfoLinks(Model):
    """Links within the model info endpoint."""

    records: URL
    """Link to the published records"""

    html: URL | None = field(default=None)
    """Link to the model records' HTML listing page"""

    drafts: Optional[URL] = field(default=None)
    """Link to the user's draft records"""

    deposit: Optional[URL] = field(default=None)
    """Deposition link for the model"""

    model: URL | None = field(default=None)
    """Link to the model definition"""
    

@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class ModelInfoContentType(Model):
    """Acceptable content-types for the model."""

    content_type: str
    """The content-type accepted by the model"""

    name: str | None = None
    """The name of the content-type"""

    description: str | None = None
    """The description of the content-type"""

    schema: URL | None = None
    """Machine parseable schema for the content-type"""
    
    can_export: bool = False
    """Whether the content-type can be used for exporting records (via Accept header)"""
    
    can_deposit: bool = False
    """Whether the content-type can be used for importing records (via Content-Type header)"""
    

@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class ModelInfo(Model):
    """Information about metadata model within invenio server."""

    type: str
    """The type of the model"""

    schema: str
    """The json schema of the model"""

    name: str
    """The name of the model"""

    description: str
    """The description of the model"""

    version: str
    """The version of the model"""

    features: list[str]
    """List of features supported by the model"""

    links: ModelInfoLinks
    """Links to the model"""

    content_types: list[ModelInfoContentType] = field(factory=list)
    """List of supported content-types for API serialization"""

    metadata: bool = field(default=True)
    """Whether the model contains metadata inside the metadata element."""


@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RepositoryInfo(Model):
    """Extra info downloaded from nrp-compatible invenio repository."""

    schema: str
    """Version of this configuration schema"""

    name: str
    """The name of the repository"""

    description: str
    """The description of the repository"""

    version: str
    """The version of the repository"""

    invenio_version: str
    """The version of invenio the repository is based on"""

    links: RepositoryInfoLinks
    """Links to the repository"""

    transfers: list[str] = field(factory=list)
    """List of supported file transfer protocols"""

    models: dict[str, ModelInfo] = field(factory=dict)
    """Information about the models in the repository"""

    default_model: str | None = field(default=None)
    """The default model for the repository. 
    If set, it is used whenever the model is not specified."""

    features: list[str] = field(factory=list)
    """List of features supported by the repository"""


#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline interface for creating records."""

from asyncio import Task, TaskGroup
from functools import partial
from typing import Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client import AsyncRecordsClient, get_async_client
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.base import set_variable as setvar
from nrp_cmd.cli.records.metadata import read_metadata
from nrp_cmd.cli.records.table_formatters import format_record_table
from nrp_cmd.config import Config
from nrp_cmd.progress import show_progress
from nrp_cmd.types.records import Record

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_progress,
    with_repository,
    with_setvar,
    with_verbosity,
)


@argument_with_help("metadata", type=str, help="Metadata")
@argument_with_help("files", type=str, nargs=-1, help="List of files to upload")
@click.option("--model", type=str, help="Model name")
@click.option("--community", type=str, help="Community name")
@click.option("--workflow", type=str, help="Workflow name")
@click.option(
    "--metadata-only/--no-metadata-only",
    default=False,
    help="The record will only have metadata",
)
@with_config
@with_output
@with_verbosity
@with_setvar
@with_repository
@with_progress
@with_model(community=True, workflow=True, draft=False, published=False)
@async_command
async def create_record(
    *,
    config: Config,
    out: Output,
    variable: Optional[str] = None,
    repository: Optional[str] = None,
    metadata: str,
    files: Optional[list[str]] = None,
    model: Model,
    metadata_only: bool = False,
) -> None:
    """Create a new record in the repository and optionally upload files to it.

    Example:
    ```
    nrp-cmd create record '{"title": "My record"}' --set-variable var
        creates the record and stores its id in the variable var

    nrp-cmd create record '{"title": "My record"}' file1.txt '{"title": "File 1"}' file2.txt '{"title": "File 2"}'
        creates the record and uploads two files to it

    nrp-cmd create record '[{...},{...},{...}]' --set-variable var
        create multiple records and store their ids in the variable var
    ```

    """
    console = Console()
    if not files:
        files = []

    if len(files) % 2 != 0:
        raise ValueError("Files must be in pairs of <path> and file metadata.")

    metadata_json = read_metadata(metadata)
    client = await get_async_client(repository, config=config)
    records_api: AsyncRecordsClient = client.records

    if model.model is not None:
        records_api = records_api.with_model(model.model)

    if not isinstance(metadata_json, list):
        assert isinstance(metadata_json, dict), "Metadata must be a dictionary."
        metadata_json = [metadata_json]

    create_results: list[Task[Record]] = []
    assert isinstance(metadata_json, list)
    with show_progress(total=len(metadata_json), quiet=not out.progress):
        async with TaskGroup() as tg:
            for record_metadata in metadata_json:
                create_results.append(
                    tg.create_task(
                        records_api.create(
                            record_metadata,
                            community=model.community,
                            workflow=model.workflow,
                            files_enabled=not metadata_only,
                            model=model.model,
                        )
                    )
                )
    records = [result.result() for result in create_results]
    if variable:
        setvar(config, variable, [str(record.links.self_) for record in records])

    # upload files - imported here to avoid circular imports
    if files:
        if len(records) > 1:
            raise ValueError("Cannot upload files to multiple records.")

        record = records[0]
        from nrp_cmd.cli.files.upload import upload_files_to_record

        to_upload = list(zip(files[::2], files[1::2], strict=False))

        with show_progress(total=len(to_upload), quiet=not out.progress):
            async with TaskGroup() as tg:
                for file_, file_metadata in to_upload:
                    tg.create_task(
                        upload_files_to_record(client, record, (file_, file_metadata))
                    )

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(format_record_table, verbosity=out.verbosity),  # type: ignore # mypy does not understand this
    ) as printer:
        if len(records) == 1:
            printer.output(records[0])
        else:
            printer.multiple()
            for rec in records:
                printer.output(rec)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Record types."""

# TODO: can not use from __future__ import annotations here because of the attrs plugin
# currently python is unable to resolve type hints for generic types
# when from __future__ import annotations is used

from typing import Any, Optional

from attrs import define, field
from yarl import URL

from ..converter import Omit, Rename, extend_serialization
from .base import Model
from .rest import BaseRecord, RESTList, RESTObjectLinks

RecordId = str | int | URL



@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class RecordLinks(RESTObjectLinks):
    """Links of a record."""

    files: Optional[URL] = None


@define(kw_only=True)
class FilesEnabled(Model):
    """Files enabled marker."""

    enabled: bool


@define(kw_only=True)
class ParentRecord(Model):
    """Parent record of the record."""

    communities: dict[str, str] | None = None
    """Communities of the record."""

    workflow: str | None = None
    """Workflow of the record."""


# extend record serialization to allow extra data and rename files to files_
@extend_serialization(Rename("files", "files_"), 
                      Omit("_etag", from_unstructure=True), 
                      allow_extra_data=True)
@define(kw_only=True)
class Record(BaseRecord):
    """Record in the repository."""

    links: RecordLinks = field()
    """Links of the record."""

    files_: Optional[FilesEnabled] = None
    """Files enabled marker."""

    parent: Optional[ParentRecord] = None

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata of the record."""
        if 'metadata' in self._extra_data:
            return self._extra_data['metadata']
        return self._extra_data


@extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class RecordList(RESTList[Record]):
    """List of records."""

    sortBy: Optional[str] = None
    """Sort by field."""
    aggregations: Optional[Any] = None
    """Aggregations."""



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line interface for getting records."""

from typing import Optional

from rich.console import Console

from nrp_cmd.async_client.connection import limit_connections
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)
from .table_formatter import (
    format_request_and_types_table,
)


@with_config
@with_repository
@with_model
@with_output
@with_verbosity
@with_resolved_vars("record_id")
@argument_with_help("record_id", type=str, help="Record ID")
@async_command
async def list_requests(
    *,
    config: Config,
    repository: Optional[str] = None,
    record_id: str,
    out: Output,
    model: Model,
) -> None:
    """Get a record from the repository."""
    console = Console()

    with limit_connections(10):
        (
            record,
            _final_record_id,
            _repository_config,
            _record_client,
            _repository_client,
        ) = await read_record(
            record_id,
            repository,
            config,
            True,
            model.model,
            model.published,
            model.draft,
        )

    with OutputWriter(
        out.output, out.output_format, console, format_request_and_types_table
    ) as printer:
        data = {
            "requests": [
                x for x in record.expanded.get("requests", [])
            ],
            "request_types": [
                x for x in record.expanded.get("request_types", [])
            ],
        }
        printer.output(data)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Bounded stream implementation."""

from typing import Any

from .base import InputStream


class BoundedStream(InputStream):
    """A stream that reads a limited amount of data from another stream."""

    def __init__(self, stream: InputStream, limit: int):
        """Initialize the stream."""
        self._stream = stream
        self._remaining = limit

    async def read(self, size: int = -1) -> bytes:
        """Read data from the stream."""
        if self._remaining <= 0:
            return b""
        if size < 0:
            size = self._remaining
        data = await self._stream.read(min(size, self._remaining))
        self._remaining -= len(data)
        return data

    def __len__(self) -> int:
        """Return the stream size."""
        return self._remaining

    def __bool__(self):
        """Return True if the stream can provide data."""
        return bool(self._remaining)

    async def close(self) -> None:
        """Close the underlying stream."""
        await self._stream.close()

    def __getattr__(self, name: str) -> Any:
        """Delegate all other calls to the underlying stream."""
        return getattr(self._stream, name)
    
    def __aiter__(self):
         return self
    
    async def __anext__(self) -> bytes:
        ret = await self.read(16384)
        if not ret:
            raise StopAsyncIteration()
        return ret
    
    


#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command-line client for invenio repositories."""

from .cli import app

__all__ = ("app",)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""User-wide configuration of repositories, usually stored in ~/.nrp/invenio-config.json."""

from __future__ import annotations

import contextlib
import json
import os
from pathlib import Path
from typing import Optional, Self

from attrs import define, field
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from yarl import URL

from ..converter import converter
from .repository import RepositoryConfig
from .variables import Variables


@define(kw_only=True)
class Config:
    """The configuration of the NRP client as stored in the configuration file."""

    repositories: list[RepositoryConfig] = field(factory=list)
    """Locally known repositories."""

    default_alias: Optional[str] = None
    """The alias of the default repository"""

    per_directory_variables: bool = True
    """Whether to load variables from a .nrp directory in the current directory.
       If set to False, the variables are loaded from the global configuration file
       located in ~/.nrp/variables.json.
    """

    _config_file_path: Optional[Path] = None
    """The path from which the config file was loaded."""

    @classmethod
    def from_file(cls, config_file_path: Optional[Path] = None) -> Self:
        """Load the configuration from a file."""
        if not config_file_path:
            if "NRP_CMD_CONFIG_PATH" in os.environ:
                config_file_path = Path(os.environ["NRP_CMD_CONFIG_PATH"])
            else:
                config_file_path = Path.home() / ".nrp" / "invenio-config.json"

        if config_file_path.exists() and config_file_path.stat().st_size > 0:
            ret = converter.structure(
                json.loads(config_file_path.read_text(encoding="utf-8")), cls
            )
        else:
            ret = cls()
        ret._config_file_path = config_file_path
        return ret

    def save(self, path: Optional[Path] = None) -> None:
        """Save the configuration to a file, creating parent directory if needed."""
        if path:
            self._config_file_path = path
        else:
            path = self._config_file_path
        assert path is not None
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(converter.unstructure(self), indent=4))

    #
    # Repository management
    #

    def get_repository(self, alias: str) -> RepositoryConfig:
        """Get a repository by its alias."""
        for repo in self.repositories:
            if repo.alias == alias:
                if not repo.enabled:
                    raise ValueError(f"Repository with alias '{alias}' is disabled")
                return repo
        raise KeyError(f"Repository with alias '{alias}' not found")

    @property
    def default_repository(self) -> RepositoryConfig:
        """Get the default repository."""
        if self.default_alias is None:
            raise ValueError("Default repository not set")
        return self.get_repository(self.default_alias)

    def add_repository(self, repository: RepositoryConfig) -> None:
        """Add a repository to the configuration."""
        self.repositories.append(repository)

    def remove_repository(self, repository: RepositoryConfig | str) -> None:
        """Remove a repository from the configuration."""
        if isinstance(repository, str):
            repository = self.get_repository(repository)
        self.repositories.remove(repository)

    def set_default_repository(self, alias: str) -> None:
        """Set the default repository by its alias."""
        try:
            next(repo for repo in self.repositories if repo.alias == alias)
        except StopIteration:
            raise ValueError(f"Repository with alias '{alias}' not found") from None
        self.default_alias = alias

    def get_repository_from_url(self, record_url: str | URL) -> RepositoryConfig:
        """Get the repository configuration for a given record URL.

        If there is no repository configuration for the given URL, a dummy
        repository configuration is returned.
        """
        record_url = URL(record_url)
        repository_root_url = record_url.with_path("/")
        for repository in self.repositories:
            if not repository.enabled:
                continue
            if repository.url == repository_root_url:
                return repository
        # return a dummy repository configuration
        return RepositoryConfig(
            alias=str(repository_root_url),
            url=repository_root_url,
            info=None,
        )

    def find_repository(self, repository: URL | str) -> RepositoryConfig:
        """Find a repository configuration by its URL or alias name."""

        # try alias first
        if isinstance(repository, str):    
            with contextlib.suppress(KeyError):
                return self.get_repository(repository)
        # if not successful, use the url to get/configure a repository
        return self.get_repository_from_url(repository)

    def load_variables(self) -> Variables:
        """Load the global variables from the configuration file."""
        if self.per_directory_variables:
            return Variables.from_file(Path.cwd() / ".nrp" / "variables.json")
        return Variables.from_file()


config_unst_hook = make_dict_unstructure_fn(
    Config, converter, _config_file_path=override(omit=True)
)
config_st_hook = make_dict_structure_fn(
    Config, converter, _config_file_path=override(omit=True)
)

converter.register_structure_hook(Config, config_st_hook)
converter.register_unstructure_hook(Config, config_unst_hook)



"""Bearer token for the invenio repository."""

import dataclasses

from yarl import URL


@dataclasses.dataclass
class BearerTokenForHost:
    """URL and bearer token for the invenio repository."""

    host_url: URL
    """URL of the repository."""

    token: str
    """Bearer token for the repository."""

    def __post_init__(self):
        """Cast the host_url to URL if it is not already."""
        if not isinstance(self.host_url, URL):
            self.host_url = URL(self.host_url)
        assert self.token is not None, "Token must be provided"




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for files."""

from asyncio import TaskGroup
from functools import partial
from pathlib import Path
from typing import Optional

from rich.console import Console

from nrp_cmd.async_client import limit_connections
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.files.table_formatters import format_files_table
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.cli.records.record_file_name import create_output_file_name
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    OutputFormat,
    VerboseLevel,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)


@with_config
@with_repository
@with_resolved_vars("record_ids")
@with_output
@with_verbosity
@with_model
@argument_with_help("record_ids", type=str, nargs=-1, help="Record ID(s)")
@async_command
async def list_files(
    *,
    config: Config,
    repository: Optional[str] = None,
    record_ids: list[str],
    out: Output,
    model: Model,
) -> None:
    """List record's files."""
    console = Console()

    with limit_connections(10):
        async with TaskGroup() as tg:
            for record_id in record_ids:
                tg.create_task(
                    list_single_record(
                        record_id,
                        console,
                        config,
                        out.output,
                        out.output_format,
                        repository,
                        model.model,
                        model.published,
                        model.draft,
                        out.verbosity,
                    )
                )


async def list_single_record(
    record_id: str,
    console: Console,
    config: Config,
    output: Optional[Path],
    output_format: Optional[OutputFormat],
    repository: Optional[str],
    model: Optional[str],
    published: bool,
    draft: bool,
    verbosity: VerboseLevel,
):
    """List and print single record files."""
    (
        record,
        _record_id_url,
        _repository_config,
        _record_client,
        repository_client,
    ) = await read_record(record_id, repository, config, False, model, published, draft)
    files = await repository_client.files.list(record)

    if output:
        output = create_output_file_name(output, str(record.id), record, output_format)
        if output.parent:
            output.parent.mkdir(parents=True, exist_ok=True)

    with OutputWriter(
        output,
        output_format,
        console,
        partial(format_files_table, record, verbosity=verbosity),
    ) as printer:
        printer.output(files)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Synchronous client for the NRP Invenio repository - low level connection."""

from .connection import ConnectionMixin, SyncConnection, connection_unstructure_hook
from .limiter import limit_connections

__all__ = (
    "SyncConnection",
    "connection_unstructure_hook",
    "ConnectionMixin",
    "limit_connections",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Protocol for data sources and sinks."""

from collections.abc import AsyncIterator
from enum import StrEnum, auto
from typing import Protocol, runtime_checkable


@runtime_checkable
class InputStream(Protocol):
    """Protocol for data readers."""

    async def read(self, n: int = -1) -> bytes:
        """Read up to n bytes from the stream."""
        ...
        
    def __aiter__(self) -> AsyncIterator[bytes]:
        """Return an async iterator returning chunks of the stream."""
        ...

    async def close(self) -> None:
        """Close the stream."""
        ...

    def __len__(self) -> int:
        """Return the length of the stream."""
        ...


@runtime_checkable
class OutputStream(Protocol):
    """Protocol for data writers."""

    async def write(self, data: bytes) -> int:
        ...

    async def close(self) -> None:
        ...


class SinkState(StrEnum):
    """State of the sink."""

    NOT_ALLOCATED = auto()
    """Sink's space has not been allocated yet."""
    ALLOCATED = auto()
    """Sink's space has been allocated and reserved on the filesystem/memory."""
    CLOSED = auto()
    """Sink has been closed."""


@runtime_checkable
class DataSink(Protocol):
    """Protocol for data sinks."""

    async def allocate(self, size: int) -> None:
        """Allocate space for the sink.

        :param size: The size of the sink in bytes.
        """
        ...

    async def open_chunk(self, offset: int = 0) -> OutputStream:
        """Get a writer for the sink, starting at the given offset.

        :param offset: The offset in bytes from the start of the sink.
        :return: A writer for the sink.
        """
        ...

    async def close(self) -> None:
        """Close the sink and all unclosed writers."""
        ...

    @property
    def state(self) -> SinkState:
        """Return the current state of the sink."""
        ...


@runtime_checkable
class DataSource(Protocol):
    """Protocol for data sources."""

    has_range_support: bool = False

    async def open(self, offset: int = 0, count: int | None = None) -> InputStream:
        """Open the data source for reading.

        :param offset:      where to start reading from
        :param count:       how many bytes to read, if None, read until the end
        :return:            a reader for the data source
        """
        ...

    async def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        """Return the checksum of the (portion of) source.

        :param algo: The checksum algorithm to use.
        :param offset: The offset in bytes from the start of the source.
        :param count: The number of bytes to include in the checksum.
        :return The checksum of the stream or None if the implementation does not support checksums.
        """
        ...

    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        ...

    async def size(self) -> int:
        """Return the length of the data source in bytes."""
        ...

    async def content_type(self) -> str:
        """Return the content type of the data source."""
        ...

    async def close(self) -> None:
        """Close the data source."""
        ...



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""File types."""

# TODO: can not use from __future__ import annotations here because of the attrs plugin
# currently python is unable to resolve type hints for generic types
# when from __future__ import annotations is used

from typing import Any, Optional

from attrs import define, field
from yarl import URL

from ..converter import Omit, Rename, extend_serialization
from .base import Model
from .rest import RESTObject, RESTObjectLinks

# some of transfer types. Repository might provide additional ones.
TRANSFER_TYPE_LOCAL = "L"
TRANSFER_TYPE_MULTIPART = "M"
TRANSFER_TYPE_FETCH = "F"
TRANSFER_TYPE_REMOTE = "R"


@extend_serialization(Rename("type", "type_"), allow_extra_data=True)
@define(kw_only=True)
class FileTransfer(Model):
    """File transfer metadata."""

    type_: str = field(default=TRANSFER_TYPE_LOCAL)


@define(kw_only=True)
class MultipartUploadLinks(Model):
    """Links for multipart uploads."""

    url: URL
    """The url where to upload the part to."""


@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class FileLinks(RESTObjectLinks):
    """Links for a single invenio file."""

    content: Optional[URL] = None
    """Link to the content of the file."""

    commit: Optional[URL] = None
    """Link to commit (finalize) uploading of the file."""

    parts: Optional[list[MultipartUploadLinks]] = None
    """For multipart upload, links where to upload the part to."""


@extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class File(RESTObject):
    """A file object as stored in .../files/<key>."""

    key: str
    """Key(filename) of the file."""

    metadata: dict[str, Any] = field(factory=dict)
    """Metadata of the file, as defined in the model."""

    links: FileLinks
    """Links to the file content and commit."""

    transfer: FileTransfer = field(
        factory=lambda: FileTransfer(type_=TRANSFER_TYPE_LOCAL.value)
    )
    """File transfer type and metadata."""

    status: Optional[str] = None

    size: Optional[int] = None


@extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class FilesList(RESTObject):
    """A list of files, as stored in ...<record_id>/files."""

    enabled: bool
    """Whether the files are enabled on the record."""

    entries: list[File] = field(factory=list)
    """List of files on the record."""

    def __getitem__(self, key: str) -> File:
        """Get a file by key."""
        for v in self.entries:
            if v.key == key:
                return v
        raise KeyError(f"File with key {key} not found")




#
# This file was generated from the asynchronous client at streams/file.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""File sources and sinks."""

import contextlib
import hashlib
from pathlib import Path
from typing import override

import magic

from .base import DataSink, DataSource, InputStream, OutputStream, SinkState
from .bounded_stream import BoundedStream
from .os import FileOutputStream, checksum_file, file_stat, open_file


class FileSink(DataSink):
    """Implementation of a sink that writes data to filesystem."""

    def __init__(self, fpath: Path):
        """Initialize the sink.

        :param fpath: The path to the file where the data will be written.
        """
        self._fpath = fpath
        self._state = SinkState.NOT_ALLOCATED
        self._file: FileOutputStream | None = None

    @override
    def allocate(self, size: int) -> None:
        """Allocate space for the sink."""
        self._file = open_file(self._fpath, mode="wb")
        self._file.truncate(size)
        self._state = SinkState.ALLOCATED

    @override
    def open_chunk(self, offset: int = 0) -> OutputStream:  # type: ignore
        """Open a chunk of the sink for writing."""
        if self._state != SinkState.ALLOCATED:
            raise RuntimeError("Sink not allocated")

        chunk = open_file(self._fpath, mode="r+b")
        chunk.seek(offset)
        return chunk

    @override
    def close(self) -> None:
        """Close the sink."""
        if self._file is not None:
            with contextlib.suppress(Exception):
                self._file.close()
        self._file = None

        self._state = SinkState.CLOSED

    @override
    @property
    def state(self) -> SinkState:
        """Return the current state of the sink."""
        return self._state

    def __repr__(self):
        """Return a string representation of the sink."""
        return f"<{self.__class__.__name__} {self._fpath} {self._state}>"


class FileSource(DataSource):
    """A data source that reads data from a file."""

    has_range_support = True

    def __init__(self, file_name: Path | str):
        """Initialize the data source.

        :param file_name: The name of the file to read from, must exist on the filesystem
        """
        if isinstance(file_name, str):
            file_name = Path(file_name)
        self._file_name = file_name

    @override
    def open(self, offset: int = 0, count: int | None = None) -> InputStream:  # type: ignore
        """Open the file for reading."""
        ret = open_file(self._file_name, mode="rb")
        ret.seek(offset)
        if not count:
            return BoundedStream(ret, self.size())
        else:
            return BoundedStream(ret, count)

    @override
    def size(self) -> int:
        """Return the size of the file."""
        return (file_stat(self._file_name)).st_size

    @override
    def content_type(self) -> str:
        """Return the content type of the file."""
        f = open_file(self._file_name, mode="rb")
        try:
            data = f.read(2048)
            return magic.from_buffer(data, mime=True)
        finally:
            f.close()

    @override
    def close(self) -> None:
        """Close the data source."""
        pass

    @override
    def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        return checksum_file(self._file_name, algo, offset, count)

    @override
    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        return list(hashlib.algorithms_available)




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Main commandline client."""

from __future__ import annotations

import dataclasses
import logging
import logging.handlers
from collections.abc import Callable
from typing import Any, Optional

import rich_click as click
from rich import box
from rich.panel import Panel
from rich.table import Table
from rich_click.rich_help_formatter import RichHelpFormatter

from nrp_cmd.cli.arguments import Argument, ClickCommand, with_errors
from nrp_cmd.cli.files import (
    delete_file,
    download_files,
    list_files,
    update_file_metadata,
    upload_files,
)
from nrp_cmd.cli.records import (
    create_record,
    delete_record,
    download_record,
    edit_record,
    get_record,
    publish_record,
    retract_record,
    scan_records,
    search_records,
    update_record,
    version_record,
)
from nrp_cmd.cli.repositories import (
    add_repository,
    describe_repository,
    disable_repository,
    enable_repository,
    list_repositories,
    remove_repository,
    select_repository,
)
from nrp_cmd.cli.repository_requests import (
    accept_request,
    cancel_request,
    create_request,
    decline_request,
    list_requests,
    submit_request,
)
from nrp_cmd.cli.variables import (
    get_variable,
    list_variables,
    remove_variable,
    set_variable,
)

logging.basicConfig(level=logging.ERROR)

commands: list[tuple[str, str, ClickCommand]] = [
    #
    #
    # verb centric
    #
    #
    ("add", "repository", add_repository),
    ("accept", "request", accept_request),
    ("cancel", "request", cancel_request),
    ("create", "record", create_record),
    ("create", "request", create_request),
    ("decline", "request", decline_request),
    ("delete", "file", delete_file),
    ("delete", "record", delete_record),
    ("describe", "repository", describe_repository),
    ("disable", "repository", disable_repository),
    ("download", "record", download_record),
    ("download", "files", download_files),
    ("download", "file", download_files),
    ("enable", "repository", enable_repository),
    ("get", "record", get_record),
    ("get", "variable", get_variable),
    ("list", "files", list_files),
    ("list", "records", search_records),
    ("list", "repositories", list_repositories),
    ("list", "requests", list_requests),
    ("list", "variables", list_variables),
    ("publish", "record", publish_record),
    ("retract", "record", retract_record),
    ("edit", "record", edit_record),
    ("version", "record", version_record),
    ("remove", "repository", remove_repository),
    ("remove", "variable", remove_variable),
    ("scan", "records", scan_records),
    ("search", "records", search_records),
    ("select", "repository", select_repository),
    ("set", "variable", set_variable),
    ("submit", "request", submit_request),
    ("upload", "file", upload_files),
    ("update", "record", update_record),
    ("update", "file", update_file_metadata),
    #
    #
    # noun centric
    #
    #
    ("files", "list", list_files),
    ("files", "delete", delete_file),
    ("files", "download", download_files),
    ("files", "upload", upload_files),
    ("files", "update", update_file_metadata),
    ("records", "create", create_record),
    ("records", "delete", delete_record),
    ("records", "download", download_record),
    ("records", "get", get_record),
    ("records", "list", search_records),
    ("records", "search", search_records),
    ("records", "scan", scan_records),
    ("records", "update", update_record),
    ("records", "edit", edit_record),
    ("records", "version", version_record),
    ("records", "publish", publish_record),
    ("records", "retract", retract_record),
    ("requests", "accept", accept_request),
    ("requests", "cancel", cancel_request),
    ("requests", "create", create_request),
    ("requests", "decline", decline_request),
    ("requests", "list", list_requests),
    ("requests", "submit", submit_request),
    ("repositories", "add", add_repository),
    ("repositories", "describe", describe_repository),
    ("repositories", "disable", disable_repository),
    ("repositories", "enable", enable_repository),
    ("repositories", "remove", remove_repository),
    ("repositories", "select", select_repository),
    ("repositories", "list", list_repositories),
    ("variables", "get", get_variable),
    ("variables", "set", set_variable),
    ("variables", "remove", remove_variable),
    ("variables", "list", list_variables),
]
"""CLI commands."""

click.rich_click.OPTION_GROUPS = {
    "nrp-cmd *": [
        {
            "name": "Output",
            "options": ["--verbose", "--quiet", "--output", "--output-format"],
        },
        {
            "name": "Configuration",
            "options": ["--config-path", "--repository"],
        },
        {
            "name": "Model",
            "options": [
                "--model",
                "--community",
                "--workflow",
                "--draft",
                "--published",
            ],
        },
        {
            "name": "Search",
            "options": ["--page", "--size", "--sort"],
        },
        {
            "name": "Variables",
            "options": ["--set"],
        },
        {
            "name": "Debugging and Logging",
            "options": [
                "--progress",
                "--log-url",
                "--log-request",
                "--log-response",
                "--log-stacktrace",
            ],
        },
    ],
}


class CommandWithAttributeHelp(click.RichCommand):
    def format_help(
        self, ctx: click.RichContext, formatter: RichHelpFormatter  # type: ignore[override]
    ) -> None:
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_arguments(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)

    def format_arguments(
        self, ctx: click.RichContext, formatter: RichHelpFormatter
    ) -> None:
        """Writes all the options into the formatter if they exist."""
        options_rows: list[tuple[str, str]] = []
        for param in self.get_params(ctx):
            if isinstance(param, Argument):
                rv = param.get_argument_help_record(ctx)
                if rv is not None:
                    options_rows.append(rv)

        if len(options_rows) > 0:
            t_styles: dict[str, Any] = {
                "show_lines": formatter.config.style_options_table_show_lines,
                "leading": formatter.config.style_options_table_leading,
                "box": formatter.config.style_options_table_box,
                "border_style": formatter.config.style_options_table_border_style,
                "row_styles": formatter.config.style_options_table_row_styles,
                "pad_edge": formatter.config.style_options_table_pad_edge,
                "padding": formatter.config.style_options_table_padding,
            }
            if isinstance(formatter.config.style_options_table_box, str):
                t_styles["box"] = getattr(box, t_styles.pop("box"), None)  # type: ignore[arg-type]

            options_table = Table(
                highlight=True,
                show_header=False,
                expand=True,
                **t_styles,  # type: ignore[arg-type]
            )
            # Strip the required column if none are required
            for row in options_rows:
                options_table.add_row(*row)

            kw: dict[str, Any] = {
                "border_style": formatter.config.style_options_panel_border,
                "title": "Arguments",
                "title_align": formatter.config.align_options_panel,
            }

            if isinstance(formatter.config.style_options_panel_box, str):
                box_style = getattr(box, formatter.config.style_options_panel_box, None)
            else:
                box_style = formatter.config.style_options_panel_box

            if box_style:
                kw["box"] = box_style

            formatter.write(Panel(options_table, **kw))


@dataclasses.dataclass
class CommandTreeNode:
    """A tree of command groups/commands."""

    children: dict[str, CommandTreeNode] = dataclasses.field(default_factory=dict)
    """Child nodes of this group."""
    command: Optional[Callable[..., None]] = None
    """Command to execute at this node, if the children are empty."""

    def register_commands(self, parent_click_group: click.Group) -> None:
        """Register the commands to parent's typer group."""
        for child_name, child in self.children.items():
            if child.children:
                children_names = ", ".join(child.children.keys())
                grp = parent_click_group.group(
                    name=child_name, help=f"{children_names}"
                )(lambda: None)
                child.register_commands(grp)
            else:
                assert child.command
                parent_click_group.command(child_name, cls=CommandWithAttributeHelp)(
                    with_errors(child.command)
                )

    def add_command(
        self,
        command_decl: tuple[Callable[..., None]]
        | tuple[str, Callable[..., None]]
        | tuple[str, str, Callable[..., None]],
    ) -> None:
        """Add a command to the tree."""
        if len(command_decl) == 1:
            command = command_decl[0]
            assert not self.command, f"Can not set {command} to node {self}"
            assert not self.children, f"Can not set {command} to node {self}"
            self.command = command
        else:
            command_name = command_decl[0]
            assert isinstance(command_name, str), f"Invalid command name {command_name}"
            if command_name not in self.children:
                self.children[command_name] = CommandTreeNode()
            self.children[command_name].add_command(
                command_decl[1:],
            )


def generate_click_command() -> click.Group:
    """Register all commands into the typer app."""
    app = click.group(name="nrp-cmd")(lambda: None)

    tree_root = CommandTreeNode()
    for cmd in commands:
        # ignore type as mypy says that the function can not be cast to Callable[..., None]
        tree_root.add_command(cmd)  # type: ignore

    tree_root.register_commands(app)

    return app


app = generate_click_command()
"""Typer main application."""


if __name__ == "__main__":
    # call the application if run as a script
    app()



#
# This file was generated from the asynchronous client at files/transfer/multipart.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Local transfer."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yarl import URL

from ...connection.aws_limits import adjust_upload_multipart_params
from . import Transfer

if TYPE_CHECKING:
    from yarl import URL

    from ....progress import ProgressBar
    from ...connection import SyncConnection
    from ...streams import DataSource
    from ..files import File, MultipartUploadLinks


class MultipartTransfer(Transfer):
    """Local transfer.

    This transfer copies a local file to the repository.
    The file will be stored in repository's primary storage (thus local)
    and the upload will be handled solely through the repository.
    """

    def prepare(
        self,
        connection: SyncConnection,
        files_link: URL,
        transfer_payload: dict,
        source: DataSource,
    ) -> None:
        """Prepare the transfer."""
        if not transfer_payload.get("size"):
            transfer_payload["size"] = source.size()

        transfer_md = transfer_payload.get("transfer", {})

        parts, part_size = adjust_upload_multipart_params(
            transfer_payload["size"],
            transfer_md.get("parts"),
            transfer_md.get("part_size"),
        )
        transfer_md["parts"] = parts
        transfer_md["part_size"] = part_size
        
    def upload(
        self,
        connection: SyncConnection,
        initialized_upload: File,
        source: DataSource,
        progress_bar: ProgressBar,
    ) -> None:
        """Upload the file."""
        links: list[MultipartUploadLinks] = initialized_upload.links.parts or []
        assert links

        number_of_parts = len(links)
        part_size: int = initialized_upload.transfer.part_size

        size = initialized_upload.size

        for pt in range(number_of_parts):
            start = pt * part_size
            count = min(part_size, size - start)

            headers: dict[str, str] = {
                "Content-Length": str(count),
                "Content-Type": "application/octet-stream",
            }

            if "md5" in source.supported_checksums():
                headers["Content-MD5"] = source.checksum(
                    "md5", offset=start, count=count
                )

            connection.put_stream(
                url=links[pt].url,
                source=source,
                open_kwargs={"offset": start, "count": count},
                headers=headers,
            )

    def get_commit_payload(self, initialized_upload: File) -> dict:
        """Get payload for finalization of the successful upload."""
        return {}




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line interface for getting records."""

import asyncio
from typing import Optional

from rich.console import Console

from nrp_cmd.async_client import (
    AsyncRecordsClient,
    get_async_client,
    get_repository_from_record_id,
)
from nrp_cmd.async_client.connection import AsyncConnection, limit_connections
from nrp_cmd.cli.base import async_command
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    with_config,
    with_model,
    with_output,
    with_record_ids,
    with_repository,
    with_verbosity,
)


@with_model
@with_config
@with_repository
@with_record_ids
@with_output
@with_verbosity
@async_command
async def delete_record(
    config: Config,
    repository: Optional[str],
    out: Output,
    record_ids: list[str],
    model: Model,
) -> None:
    """Get a record from the repository."""
    console = Console()

    with limit_connections(10):
        async with asyncio.TaskGroup() as tg:
            for record_id in record_ids:
                tg.create_task(
                    delete_single_record(
                        record_id,
                        console,
                        config,
                        repository,
                        model.model,
                        model.published,
                        model.draft,
                        out.verbosity,
                    )
                )


async def delete_single_record(
    record_id: str,
    console: Console,
    config: Config,
    repository: str | None,
    model: str | None,
    published: bool,
    draft: bool,
    verbosity: VerboseLevel,
) -> None:
    """Get a single record from the repository and print/save it."""
    connection = AsyncConnection()
    
    final_record_id, repository_config = await get_repository_from_record_id(
        connection, record_id, config, repository
    )
    # set it temporarily to the config
    config.add_repository(repository_config)
    client = await get_async_client(repository, config=config)
    records_api: AsyncRecordsClient = client.records
    if model is not None:
        records_api = records_api.with_model(model)
    if published:
        records_api = records_api.published_records
    if draft:
        records_api = records_api.draft_records

    await records_api.delete(final_record_id)
    if verbosity != VerboseLevel.QUIET:
        console.print(f"[green]Record with id {record_id} has been deleted.[/green]")



#
# This file was generated from the asynchronous client at streams/progress.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


from collections.abc import Iterator
from typing import override

from ...progress import ProgressBar
from .base import DataSink, DataSource, InputStream, OutputStream, SinkState


class ProgressIterator(Iterator[bytes]):
    """Iterator with progress bar."""

    def __init__(self, iterator: Iterator[bytes], source: "ProgressSource"):
        """Initialize the iterator."""
        self._iterator = iterator
        self._source = source

    @override
    def __next__(self) -> bytes:
        data = next(self._iterator)
        self._source.update_progress(len(data))
        return data


class ProgressInputStream(InputStream):
    """Input stream with progress bar."""

    def __init__(self, stream: InputStream, source: "ProgressSource"):
        """Initialize the input stream."""
        self._stream = stream
        self._source = source

    @override
    def read(self, n: int = -1) -> bytes:
        data = self._stream.read(n)
        self._source.update_progress(len(data))
        return data

    @override
    def __iter__(self) -> Iterator[bytes]:
        return ProgressIterator(aiter(self._stream), self._source)

    @override
    def __len__(self) -> int:
        return len(self._stream)

    @override
    def close(self) -> None:
        return self._stream.close()


class ProgressSource(DataSource):
    """Data source with progress bar."""

    def __init__(self, source: DataSource, progress_bar: ProgressBar):
        """Create the data source."""
        self._source = source
        self._progress_bar = progress_bar
        self._output_progress = 0

    @property
    def has_range_support(self) -> bool:
        """Return whether the source supports range requests."""
        return self._source.has_range_support

    @has_range_support.setter
    def has_range_support(self, value: bool) -> None:
        """Set whether the source supports range requests."""
        raise AttributeError("Cannot set has_range_support on ProgressSource")

    @override
    def open(self, offset: int = 0, count: int | None = None) -> InputStream:
        # reset the progress bar
        self._progress_bar.increment(-self._output_progress)
        self._output_progress = 0
        return ProgressInputStream(self._source.open(offset, count), self)

    @override
    def size(self) -> int:
        return self._source.size()

    @override
    def content_type(self) -> str:
        return self._source.content_type()

    @override
    def close(self) -> None:
        return self._source.close()

    def update_progress(self, count: int) -> None:
        """Update the progress bar."""
        self._output_progress += count
        self._progress_bar.increment(count)

    @override
    def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        return self._source.checksum(algo, offset, count)

    @override
    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        return self._source.supported_checksums()


class ProgressSink(DataSink):
    """Data sink with progress bar."""

    def __init__(self, sink: DataSink, progress_bar: ProgressBar):
        """Create the data sink."""
        self._sink = sink
        self._progress_bar = progress_bar
        self._input_progress = 0

    @override
    def allocate(self, size: int) -> None:
        return self._sink.allocate(size)

    @override
    def open_chunk(self, offset: int = 0) -> OutputStream:
        self._progress_bar.increment(-self._input_progress)
        self._input_progress = 0
        return ProgressOutputStream(self._sink.open_chunk(offset), self)

    @override
    def close(self) -> None:
        return self._sink.close()

    @override
    @property
    def state(self) -> SinkState:
        return self._sink.state
    
    def update_progress(self, count: int) -> None:
        """Update the progress bar."""
        self._input_progress += count
        self._progress_bar.increment(count)


class ProgressOutputStream(OutputStream):
    """Output stream reporting progress."""
    
    def __init__(self, stream: OutputStream, sink: ProgressSink) -> None:
        """Initialize the output stream."""
        super().__init__()
        self._stream = stream
        self._sink = sink
    
    @override
    def write(self, data: bytes) -> int:
        ret = self._stream.write(data)
        self._sink.update_progress(ret)
        return ret

    @override
    def close(self) -> None:
        self._stream.close()




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line client for editing published records."""

import asyncio
from functools import partial
from pathlib import Path
from typing import Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client.connection import limit_connections
from nrp_cmd.cli.base import OutputFormat, OutputWriter, async_command
from nrp_cmd.cli.records.record_file_name import create_output_file_name
from nrp_cmd.cli.records.table_formatters import format_record_table
from nrp_cmd.config import Config
from nrp_cmd.errors import DoesNotExistError
from nrp_cmd.types.records import Record
from nrp_cmd.types.requests import Request

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    with_config,
    with_model,
    with_output,
    with_record_ids,
    with_repository,
    with_verbosity,
)
from ..repository_requests.table_formatter import format_request_table
from .get import read_record


@with_config
@with_repository
@with_record_ids
@with_verbosity
@with_output
@with_model(draft=False, published=False)
@async_command
async def edit_record(
    config: Config,
    repository: Optional[str],
    record_ids: list[str],
    model: Model,
    out: Output,
) -> None:
    """Edit published record's metadata."""
    console = Console()

    with limit_connections(10):
        tasks: list[asyncio.Task[Record | Request | None]] = []
        async with asyncio.TaskGroup() as tg:
            for record_id in record_ids:
                tasks.append(
                    tg.create_task(
                        edit_single_record(
                            record_id,
                            console,
                            config,
                            repository,
                            model.model,
                            out.output,
                            out.output_format,
                            out.verbosity,
                        )
                    )
                )
        results = [x.result() for x in tasks]
        for r in results:
            if r is None:
                raise click.Abort()


async def edit_single_record(
    record_id: str,
    console: Console,
    config: Config,
    repository: str | None = None,
    model: str | None = None,
    output: Path | None = None,
    output_format: OutputFormat | None = None,
    verbosity: VerboseLevel = VerboseLevel.NORMAL,
) -> Record | Request | None:
    """Publish a record."""
    try:
        (
            record,
            _final_record_id,
            _repository_config,
            record_client,
            _repository_client,
        ) = await read_record(
            record_id, repository, config, False, model, published=False, draft=True
        )
    except DoesNotExistError as e:
        console.print(f"[red]Record with id {record_id} does not exist.[/red]")
        if verbosity == VerboseLevel.VERBOSE:
            print(e)
        return None

    ret = await record_client.edit_metadata(record)

    if output:
        output = create_output_file_name(
            output, str(record.id or record_id or "unknown_id"), record, output_format
        )
    if output and output.parent:
        output.parent.mkdir(parents=True, exist_ok=True)

    # note: this is synchronous, but it is not a problem as only metadata are printed/saved
    if isinstance(ret, Record):
        with OutputWriter(
            output,
            output_format,
            console,
            partial(format_record_table, verbosity=verbosity),  # type: ignore # mypy does not understand this
        ) as printer:
            printer.output(ret)
    else:
        with OutputWriter(
            output,
            output_format,
            console,
            partial(format_request_table, verbosity=verbosity),  # type: ignore # mypy does not understand this
        ) as printer:
            printer.output(ret)

    return ret



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Contains the limiter class for limiting the number of simultaneous connections."""

import asyncio
import contextlib
import contextvars
from collections.abc import AsyncGenerator, Generator

from yarl import URL


class Limiter(asyncio.Semaphore):
    """A class to limit the number of simultaneous connections."""

    def __init__(self, capacity: int, per_server_capacity: int = -1):
        """Initialize the limiter.

        :param capacity:    the number of simultaneous connections
        """
        if capacity <= 0:
            capacity = 10
        self.capacity = capacity
        super().__init__(capacity)

    @property
    def free(self) -> int:
        """The number of free slots.

        :return:   the number of remaining connections (approximate)
        """
        return self._value

    @contextlib.asynccontextmanager
    async def limit(self, url: URL) -> AsyncGenerator[None, None]:
        async with self:
            yield


current_limiter_var = contextvars.ContextVar[Limiter]("current_limiter")


class CurrentLimiterProxy:
    """A class to manage the current limiter."""

    @property
    def free(self) -> int:
        """The number of free slots.

        :return:   the number of remaining connections (approximate)
        """
        return current_limiter_var.get().free

    @contextlib.asynccontextmanager
    async def limit(self, url: URL) -> AsyncGenerator[None, None]:
        try:
            limiter = current_limiter_var.get()
        except LookupError:
            limiter = None
        if limiter is None:
            limiter = Limiter(10)
            current_limiter_var.set(limiter)
        async with limiter.limit(url):
            yield

    def reset(self):
        """Reset the current limiter."""
        current_limiter_var.set(None)

current_limiter = CurrentLimiterProxy()

@contextlib.contextmanager
def limit_connections(
    max_connections: int, per_server_connections: int = -1
) -> Generator[None, None, None]:
    """Limit the number of simultaneous connections."""
    token = current_limiter_var.set(Limiter(max_connections, per_server_connections))
    try:
        yield
    finally:
        current_limiter_var.reset(token)


#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Synchronous task group."""

from collections.abc import Callable


class TaskResult[T]:
    """Synchronous version of TaskResult."""

    def __init__(self, result: T):
        self._result = result

    def result(self) -> T:
        return self._result


class TaskGroup:
    """Synchronous version of TaskGroup."""

    def create_task[_T](  # type: ignore
        self,
        task: Callable[[], _T],
    ) -> TaskResult[_T]:
        return TaskResult(task())
        
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

Task = None

__all__ = ("TaskGroup", "Task")



"""Progress bar for the command line."""
from __future__ import annotations

import contextlib
import contextvars
from threading import RLock
from typing import TYPE_CHECKING, Any, Protocol, override

from tqdm import tqdm

if TYPE_CHECKING:
    from collections.abc import Generator
    


class DummyLock:
    """A dummy lock that does nothing."""

    def __enter__(self):
        """Enter the lock."""
        yield self
    
    def __exit__(self, *args: Any):
        """Exit the lock."""
        pass


class ProgressBar(Protocol):
    """An abstract interface for a single progress bar."""
    
    def increment(self, progress: int) -> None:
        """Increment the progress by the given value."""
        ...
    
    def set_value(self, progress: int) -> None:
        """Set the progress to the given value."""
        ...
    
    def set_total(self, total_value: int) -> None:
        """Set the total value of the progress."""
        ...

    def finish(self) -> None:
        """Finish the progress."""
        ...

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        ...


class Progress(Protocol):
    """A progress that can have multiple sub-progresses.
    
    Some of the sub-processes can be short tasks that do not show a progress bar, 
    but just increase the size of the primary progress bar by 1. 
    
    Some are long tasks that show a separate progress bar.
    
    ```
    p = TQDMProgress()
    
    with p.long_task("Doing something big") as progress:
        p.set_total(100)
        for i in range(100):
            progress.update(1)
            
    with p.short_task() as progress:
        time.sleep(1)
    ```
    
    """

    def increment(self, progress: int) -> None:
        """Increment the progress by the given value."""
        ...
    
    def set_value(self, progress: int) -> None:
        """Set the progress to the given value."""
        ...

    def set_total(self, total_value: int) -> None:
        """Set the total value of the progress."""
        ...

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        ...

    @contextlib.contextmanager
    def long_task(self, name: str, increment_total: bool=True) -> Generator[ProgressBar, None, None]:
        """Start a long task with the given name.
        
        The long task will be represented by a tqdm secondary progress bar.
        As this is a context manager, you do not need to call `finish` on the
        returned progress object.
        """
        progress = self.start_long_task(name, increment_total)
        try:
            yield progress
        finally:
            progress.finish()
        
        
    @contextlib.contextmanager
    def short_task(
        self, increment_total: bool = False
    ) -> Generator[ProgressBar, None, None]:
        """Start a short task.
        
        The short task will not show a progress bar,
        instead it will increase the size of the primary progress bar by 1
        and move the one tick when the task is finished.
        """
        progress = self.start_short_task(increment_total)
        try:
            yield progress
        finally:
            progress.finish()

    def start_short_task(self, increment_total: bool = False) -> ProgressBar:
        """Start a short task.
        
        Mostly it is better to use the `short_task` context manager. If you start
        the task manually, you need to call `finish` on the returned progress object.
        """
        ...

    def start_long_task(self, name: str, increment_total: bool = True) -> ProgressBar:
        """Start a long task with the given name.
        
        Mostly it is better to use the `long_task` context manager. If you start
        the task manually, you need to call `finish` on the returned progress object.
        """
        ...
    
    def finish(self) -> None:
        """Finish and remove the progress."""
        pass
    
    
class DummyProgressBar(ProgressBar):
    """A dummy progress bar that does nothing."""
    
    @override
    def increment(self, progress: int) -> None:
        pass
    
    @override
    def set_value(self, progress: int) -> None:
        pass
    
    @override
    def set_total(self, total_value: int) -> None:
        pass

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        pass

    @override
    def finish(self) -> None:
        pass
    
    
class DummyProgress(Progress):
    """A dummy progress that does nothing."""

    @override
    def increment(self, progress: int) -> None:
        pass
    
    @override
    def set_value(self, progress: int) -> None:
        pass

    @override
    def set_total(self, total_value: int) -> None:
        pass

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        pass

    @override
    def start_short_task(self, increment_total: bool = False) -> ProgressBar:
        return DummyProgressBar()

    @override
    def start_long_task(self, name: str, increment_total: bool = True) -> ProgressBar:
        return DummyProgressBar()

    @override
    def finish(self) -> None:
        pass


class TQDMProgressBar(ProgressBar):
    """A progress bar that uses tqdm to show progress."""
    
    def __init__(self, progress: TQDMProgress, desc: str, position: int = 0):
        """Create a new progress bar."""
        self._progress = progress
        self._position = position
        self.bar = tqdm(total=1, desc=desc, position=position, delay=0, leave=False)
        
    @override
    def increment(self, progress: int) -> None:
        self.bar.update(progress)
        with self.bar.get_lock():
            self._progress.increment(progress)
    
    @override
    def set_value(self, progress: int) -> None:
        with self.bar.get_lock():
            self.increment(progress - self.bar.n)
    
    @override
    def set_total(self, total_value: int) -> None:
        previous_total = self.bar.total
        self.bar.total = total_value
        self.bar.update(0)
        self._progress._change_total(total_value - previous_total)

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        self.set_total(self.bar.total + total_increment)

    @override
    def finish(self) -> None:
        self.bar.close()
        self._progress._finish_bar(self._position)            
        
        
class TQDMShortProgressBar(DummyProgressBar):
    """A short progress bar that does not show a progress bar, but just increments the total."""

    def __init__(self, progress: TQDMProgress):
        """Create a new short progress bar."""
        self._progress = progress
        
    @override
    def finish(self) -> None:
        self._progress.increment(1)
        
        
class TQDMProgress(Progress):
    """A progress that uses tqdm to show progress bars.
    
    Note: TQDM is not thread safe, so in synchronous code that can use
    threading to do parallel downloads/uploads, you should set `thread_safe=True`
    to use a `threading.RLock` to protect the tqdm instance.
    """

    def __init__(self, thread_safe: bool=False, total: int=0):
        """Create a new progress."""
        self.lock = DummyLock() if not thread_safe else RLock()
        self.bar = tqdm(total=total, position=0, leave=False)
        # initialize the internal bar's write lock
        self.bar.get_lock()
        self.used_positions: set[int] = set()
        
    @override
    def start_short_task(self, increment_total: bool = False) -> ProgressBar:
        if increment_total:
            self.bar.total += 1
            self.bar.update(0)
        progress = TQDMShortProgressBar(self)
        return progress

    @override
    def start_long_task(self, name: str, increment_total: bool = True) -> ProgressBar:
        if increment_total:
            self.bar.total += 1
            self.bar.update(0)
        progress = TQDMProgressBar(self, name, self._get_unused_position())
        return progress
    
    @override
    def increment(self, progress: int) -> None:
        # this can be called from sub-progresses, so we need to lock
        with self.lock:
            self.bar.update(progress)
    
    @override
    def set_value(self, progress: int) -> None:
        # this can be called from sub-progresses, so we need to lock
        with self.lock:
            self.bar.update(progress - self.bar.n)
    
    @override
    def set_total(self, total_value: int) -> None:
        # this can be called from sub-progresses, so we need to lock
        with self.lock:
            self.bar.total = total_value
            self.bar.update(0)

    def increment_total(self, total_increment: int) -> None:
        """Increment the total value of the progress."""
        self.set_total(self.bar.total + total_increment)

    @override
    def finish(self) -> None:
        """Finish and remove the progress."""
        self.bar.close()

    def _finish_bar(self, position: int) -> None:
        with self.lock:
            self.increment(1)
            self.used_positions.remove(position)

    def _get_unused_position(self) -> int:
        with self.lock:
            i = 1
            while True:
                if i not in self.used_positions:
                    self.used_positions.add(i)
                    return i
                i += 1
                
    def _change_total(self, change: int) -> None:
        with self.lock:
            self.bar.total += change
            self.bar.update(0)

current_progress_container = contextvars.ContextVar[Progress](
    "current_progress", default=DummyProgress()
)


class CurrentProgress(Progress):
    """A progress that uses the current progress object."""

    @override
    def increment(self, progress: int) -> None:
        current_progress_container.get().increment(progress)

    @override
    def set_value(self, progress: int) -> None:
        current_progress_container.get().set_value(progress)

    @override
    def set_total(self, total_value: int) -> None:
        current_progress_container.get().set_total(total_value)

    @override
    def increment_total(self, total_value: int) -> None:
        current_progress_container.get().increment_total(total_value)

    @override
    def start_short_task(self, increment_total: bool = False) -> ProgressBar:
        return current_progress_container.get().start_short_task(increment_total)

    @override
    def start_long_task(self, name: str, increment_total: bool = True) -> ProgressBar:
        return current_progress_container.get().start_long_task(name, increment_total)

    @override
    def finish(self) -> None:
        current_progress_container.get().finish()


current_progress = CurrentProgress()


@contextlib.contextmanager
def show_progress(
    *, impl_class: type[Progress] | None = None, total: int = 0, quiet: bool = False
) -> Generator[Progress]:
    """Create a progress object using the given implementation class."""
    if not impl_class:
        impl_class = DummyProgress if quiet else TQDMProgress
    progress = impl_class()
    token = current_progress_container.set(progress)
    try:
        if total:
            progress.set_total(total)
        yield progress
    finally:
        progress.finish()
        current_progress_container.reset(token)


if __name__ == "__main__":
    import asyncio
    
    async def main() -> None:
        """Just an example of how to use the progress bars."""
        i: int
        
        progress = TQDMProgress()
        progress.set_total(20)
        for _ in range(20):
            progress.increment(1)
            await asyncio.sleep(0.1)
        progress.finish()
    
        progress = TQDMProgress(total=20)
        for _ in range(20):
            with progress.short_task(increment_total=False):
                await asyncio.sleep(0.1)
        progress.finish()
    
        progress = TQDMProgress()
        progress.set_total(10)
        for i in range(10):
            with progress.long_task(f"Task {i}", increment_total=False) as long_progress:
                long_progress.set_total(10)
                for _ in range(10):
                    long_progress.increment(1)
                    await asyncio.sleep(0.1)
        progress.finish()

        progress = TQDMProgress()
        async with asyncio.TaskGroup() as progress_group:
            for i in range(10):
                async def worker(i: int) -> None:
                    with progress.long_task(f"Task {i}") as long_progress:
                        long_progress.set_total(20)
                        for _ in range(20):
                            long_progress.increment(1)
                            await asyncio.sleep(0.1)
                progress_group.create_task(worker(i))
        progress.finish()
    
    asyncio.run(main())


from pathlib import Path
from typing import Any, overload, override

from yarl import URL

from ...progress import DummyProgressBar, current_progress
from ...types.files import TRANSFER_TYPE_LOCAL, File, FilesList
from ...types.info import RepositoryInfo
from ...types.records import Record
from ..base_client import AsyncFilesClient
from ..connection import AsyncConnection
from ..streams import DataSink, DataSource


class AsyncInvenioFilesClient(AsyncFilesClient):
    """Invenio files client."""

    def __init__(self, connection: AsyncConnection, info: RepositoryInfo):
        """Initialize the client."""
        self._connection = connection
        self._info = info

    @override
    async def list(
        self,
        record_or_url: Record | URL,
    ) -> list[File]:
        """List the files of a record."""
        files_url = self._get_files_url(record_or_url)
        files_list = await self._connection.get(url=files_url, result_class=FilesList)
        return files_list.entries

    @override
    @overload
    async def read(self, file_url: URL) -> File: ...

    @override
    @overload
    async def read(self, record: Record, key: str) -> File: ...

    @override
    @overload
    async def read(self, record_url: URL, key: str) -> File: ...

    @override  # type: ignore
    async def read(
        self,
        *args: Record | URL | str,
    ):
        if len(args) == 1:
            assert isinstance(args[0], URL), "Invalid argument - expecting a file URL"
            file_url: URL = args[0]
        elif len(args) == 2:
            assert isinstance(
                args[0], Record
            ), "Invalid argument - expecting a record as the first argument"
            record_or_url: Record | URL = args[0]
            assert isinstance(
                args[1], str
            ), "Invalid argument - expecting a string key as the second argument"
            key: str = args[1]

            files_url = self._get_files_url(record_or_url)
            file_url = files_url / key

        return await self._connection.get(url=file_url, result_class=File)

    @override
    async def update(self, file: File) -> File:
        """Update the file metadata in the repository.

        :param file: file to update
        """
        return await self._connection.put(
            url=file.links.self_, json={"metadata": file.metadata}, result_class=File
        )

    @override
    async def upload(
        self,
        record_or_url: Record | URL,
        key: str,
        metadata: dict[str, Any],
        source: DataSource | str | Path,
        transfer_type: str = TRANSFER_TYPE_LOCAL,
        transfer_metadata: dict | None = None,
        progress: str | None = None,
    ) -> File:
        """Upload a file to the repository.

        :param record_or_url: record or url of the record where the file will be uploaded
        :param key: key of the file
        :param file: file to upload
        :param metadata: metadata of the file
        """
        files_url = self._get_files_url(record_or_url)

        if isinstance(source, (str, Path)):
            from ..streams import FileSource

            source = FileSource(source)

        # 1. initialize the upload
        transfer_md: dict[str, Any] = {}
        transfer_payload = {"key": key, "metadata": metadata, "transfer": transfer_md}
        if transfer_type != TRANSFER_TYPE_LOCAL:
            transfer_md["type"] = transfer_type

        if transfer_metadata:
            transfer_md.update(transfer_metadata)

        from .transfer import transfer_registry

        transfer = transfer_registry.get(transfer_type)

        await transfer.prepare(self._connection, files_url, transfer_payload, source)

        with current_progress.short_task():
            initialized_upload: FilesList = await self._connection.post(
                url=files_url,
                json=[transfer_payload],
                result_class=FilesList,
            )

        initialized_upload_metadata = initialized_upload[key]

        # 2. upload the file using one of the transfer types
        if progress:
            progress_bar = current_progress.start_long_task(progress)
        else:
            progress_bar = DummyProgressBar()
        progress_bar.set_total(await source.size())
        try:
            await transfer.upload(
                self._connection, initialized_upload_metadata, source, progress_bar
            )
        finally:
            progress_bar.finish()

        # 3. prepare the commit payload
        commit_payload = await transfer.get_commit_payload(initialized_upload_metadata)

        with current_progress.short_task():
            if initialized_upload_metadata.links.commit:
                committed_upload = await self._connection.post(
                    url=initialized_upload_metadata.links.commit,
                    json=commit_payload,
                    result_class=File,
                )

                return committed_upload
            else:
                return initialized_upload_metadata

    @override
    @overload
    async def download(
        self,
        record: Record,
        key: str,
        sink: DataSink,
        *,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None = None,
    ) -> None: ...

    @override
    @overload
    async def download(
        self,
        file_or_url: File | URL,
        sink: DataSink,
        *,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None = None,
    ) -> None: ...

    @override  # type: ignore
    async def download(
        self,
        *args: Record | str | DataSink | File | URL,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None = None,
    ) -> None:
        """Download the file to the sink.

        :param file_rec: file to download
        :param sink: sink where to download the file
        :param parts: number of parts to download the file in
        :param part_size: size of the parts
        :param progress: progress message in progress bar
        """
        sink: DataSink
        if len(args) == 2:
            assert isinstance(
                args[0], (File, URL)
            ), "Invalid arguments - expecting a File or URL as the first argument"
            file_or_url: File | URL = args[0]
            assert isinstance(
                args[1], DataSink
            ), "Invalid arguments - expecting a DataSink as the second argument"
            sink = args[1]
        elif len(args) == 3:
            assert isinstance(
                args[0], Record
            ), "Invalid arguments - expecting a Record as the first argument when passing file key"
            record: Record = args[0]
            assert isinstance(
                args[1], str
            ), "Invalid arguments - expecting a string key as the second argument"
            key: str = args[1]
            assert isinstance(
                args[2], DataSink
            ), "Invalid arguments - expecting a DataSink as the third argument"
            sink = args[2]
            file_or_url = await self.read(record, key)
        else:
            raise ValueError(
                "Invalid arguments - expecting either (File, DataSink), (FileURL, DataSink) or (Record, key, DataSink)"
            )

        if not isinstance(file_or_url, File):
            file_or_url = await self.read(file_or_url)

        content_url = file_or_url.links.content
        if not content_url:
            raise ValueError("The file does not have a content link")

        if progress:
            progress_bar = current_progress.start_long_task(progress)
        else:
            progress_bar = DummyProgressBar()

        await self._connection.download_file(
            content_url, sink, parts, part_size, progress_bar
        )

    def _get_files_url(self, record_or_url: Record | URL) -> URL:
        """Get the files url from the record or url."""
        if isinstance(record_or_url, Record):
            if not record_or_url.links.files:
                raise ValueError("The record does not have a files link, probably files are not enabled on it")
            return record_or_url.links.files
        return record_or_url

    @overload
    async def delete(self, record: Record, key: str | None = None) -> None: ...

    @overload
    async def delete(self, file: File) -> None: ...

    @overload
    async def delete(self, file: URL) -> None: ...

    async def delete(  # type: ignore
        self,
        arg: Record | File | URL,
        key: str | None = None,
    ) -> None:
        """Delete a record inside the repository.

        Params:
          - record and key
          - File object
          - URL of the file
        """
        if isinstance(arg, Record):
            files_url = arg.links.files
            if files_url is None:
                raise ValueError("The record does not have a files link")
            if key is None:
                raise ValueError("The key of the file to delete must be provided")
            file_url = files_url / key
        elif isinstance(arg, File):
            file_url = arg.links.self_
        else:
            file_url = arg

        await self._connection.delete(url=file_url)



#
# This file was generated from the asynchronous client at streams/os.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Low-level compatibility layer for file operations."""

import base64
import hashlib
import os
from pathlib import Path
from typing import Literal, overload

import aiofiles
import aiofiles.os

from .base import InputStream, OutputStream


class FileInputStream(InputStream):
    def seek(self, offset: int, whence: int = os.SEEK_SET) -> None:
        """Change the stream position."""
        ...
        
class FileOutputStream(OutputStream):
    def seek(self, offset: int, whence: int = os.SEEK_SET) -> None:
        """Change the stream position."""
        ...

    def truncate(self, size: int) -> None:
        ...
        
@overload
def open_file(_fpath: Path, mode: Literal["rb"]) -> FileInputStream:
    ...

@overload
def open_file(_fpath: Path, mode: Literal["wb"] | Literal["r+b"]) -> FileOutputStream:
    ...

def open_file(_fpath: Path, mode: Literal["rb"] | Literal["wb"] | Literal["r+b"]) -> FileInputStream | FileOutputStream:
    """Open a file for reading or writing."""
    r: FileInputStream | FileOutputStream = aiofiles.open(_fpath, mode=mode)  # noqa # type: ignore
    return r


def file_stat(_fpath: Path) -> os.stat_result:
    """Get file statistics."""
    return aiofiles.os.stat(_fpath)

def checksum_file(
    file_name: Path, algo: str = "md5", offset: int = 0, count: int | None = None
) -> str:
    """Calculate the checksum of the file."""
    # cpu-bound operation - run in a separate thread synchronously
    with open(file_name, mode="rb") as f:
        if offset > 0:
            f.seek(offset)

        hasher = hashlib.new(algo)
        chunk_size = get_filesystem_block_size(file_name) * 16

        while count != 0:
            chunk = f.read(min(chunk_size, count or chunk_size))
            if not chunk:
                break
            hasher.update(chunk)
            if count is not None:
                count -= len(chunk)

        return base64.b64encode(hasher.digest()).decode("ascii")


def get_filesystem_block_size(file_path: Path) -> int:
    if hasattr(os, "statvfs"):  # Unix-like systems
        statvfs = os.statvfs(str(file_path))
        return statvfs.f_bsize
    else:
        return 4096  # default block size for Windows


__all__ = (
    "open_file",
    "file_stat",
    "FileInputStream",
    "FileOutputStream",
    "checksum_file",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for downloading files."""
# TODO: test !!!!
import sys
from asyncio import Task, TaskGroup
from pathlib import Path
from typing import Any, Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client.streams import FileSink
from nrp_cmd.cli.base import OutputFormat, async_command
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.cli.records.record_file_name import create_output_file_name
from nrp_cmd.config import Config
from nrp_cmd.converter import converter
from nrp_cmd.progress import show_progress

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_progress,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_progress
@with_verbosity
@with_model
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("keys", type=str, nargs=-1, help="File key")
@click.option("-o", "--output", type=click.Path(), help="Output path")
@async_command
async def download_files(
    *,
    config: Config,
    repository: Optional[str] = None,
    record_id: str,
    keys: list[str],
    output: Optional[Path] = None,
    model: Model,
    out: Output,
) -> None:
    """Download files from a record."""
    output = output or Path.cwd()
    console = Console()

    with show_progress(total=1, quiet=not out.progress):
        await download_single_record_files(
            console,
            record_id,
            keys,
            output,
            config,
            repository,
            model.model,
            model.published,
            model.draft,
            verbosity=out.verbosity,
        )


async def download_single_record_files(
    console: Console,
    record_id: str,
    keys: list[str],
    output: Path,
    config: Config,
    repository: str | None,
    model: str | None,
    published: bool,
    draft: bool,
    verbosity: VerboseLevel,
):
    (
        record,
        record_id_url,
        _repository_config,
        _record_client,
        repository_client,
    ) = await read_record(record_id, repository, config, False, model, published, draft)
    assert repository_client
    file_client = repository_client.files

    files = await file_client.list(record)

    if "*" in keys:
        keys = [file.key for file in files]
    else:
        allowed_keys = {file.key for file in files}
        if set(keys) - allowed_keys:
            console.print(
                f"[red]Some keys not found in the record: {set(keys) - allowed_keys}[/red]"
            )
            return False

    tasks: list[tuple[str, Any, Task[Any]]] = []
    async with TaskGroup() as tg:
        for key in keys:
            try:
                file_ = next(f for f in files if f.key == key)
            except KeyError:
                print(f"Key {key} not found in files, skipping ...", file=sys.stderr)
                continue

            is_file = "{key}" in str(output)

            # sanitize the key
            if "/" in key:
                key = key.replace("/", "_")
            if ":" in key:
                key = key.replace(":", "_")

            file_output = create_output_file_name(
                output,
                key,
                file_,
                OutputFormat.JSON,
                record=converter.unstructure(record),  # type: ignore
            )

            if not is_file:
                file_output = file_output / key

            if file_output and file_output.parent:
                file_output.parent.mkdir(parents=True, exist_ok=True)

            tasks.append((
                key,
                file_output,
                tg.create_task(
                    file_client.download(
                        file_, FileSink(file_output), progress=file_.key
                    )
                ),
            ))

    ok = True
    for key, fname, task in tasks:
        if task.exception():
            ok = False
            console.print(
                f"[red]Failed to download file {key} of {record_id_url}[/red]"
            )
        else:
            if verbosity != VerboseLevel.QUIET:
                console.print(
                    f"[green]File {key} of {record_id_url} downloaded to {fname}[/green]"
                )
    return ok



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Resolve DOIs to URLs."""

from yarl import URL

from .connection import AsyncConnection


async def resolve_doi(connection: AsyncConnection, doi: str) -> str:
    """Resolve a DOI to a record URL."""
    data = await connection.get(url=URL(f"https://api.datacite.org/dois/{doi}"), result_class=dict)
    return data["data"]["attributes"]["url"]



#
# This file was generated from the asynchronous client at streams/stdin.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Data source that reads data from standard input."""

from pathlib import Path

from .base import DataSource, InputStream
from .os import open_file


class StdInDataSource(DataSource):
    """A data source that reads data from standard input."""

    def __init__(self) -> None:
        super().__init__()
        self._opened = False

    def open(self, offset: int = 0, count: int | None = None) -> InputStream:
        """Open the data source for reading."""
        if self._opened:
            raise RuntimeError("Cannot open the same data source multiple times.")
        self._opened = True
        if count is not None:
            raise ValueError("Cannot read a bounded stream from standard input.")
        if offset != 0:
            raise ValueError("Cannot seek in standard input.")
        ret = open_file(Path("/sys/stdin"), mode="rb")
        return ret

    def size(self) -> int:
        """Return the size of the data - in this case -1 as unknown."""
        return -1

    def content_type(self) -> str:
        """Return the content type of the data."""
        return "application/octet-stream"

    def close(self) -> None:
        """Close the data source."""
        pass

    def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        raise NotImplementedError("Checksums are not supported for standard input.")

    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        return []




"""Types used inside the nrp client."""

from .base import Model
from .info import (
    ModelInfo,
    ModelInfoContentType,
    ModelInfoLinks,
    RepositoryInfo,
    RepositoryInfoLinks,
)
from .records import (
    FilesEnabled,
    ParentRecord,
    Record,
    RecordId,
    RecordLinks,
    RecordList,
)

__all__ = (
    "Model", "RepositoryInfo", "RepositoryInfoLinks", "ModelInfo", "ModelInfoContentType", "ModelInfoLinks",
    "Record", "RecordLinks", "FilesEnabled", "ParentRecord", "RecordList", "RecordId"
)


#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Bearer authentication support for aiohttp."""

from typing import Optional

from aiohttp import BasicAuth, ClientRequest, hdrs

from ...types.auth import BearerTokenForHost


class Authentication(BasicAuth):
    """A generic authentication class that can be used to provide different types of authentication."""

    # a little strange to inherit from BasicAuth, but it is the only way to
    # provide different auth types to the ClientRequest

    def apply(self, request: ClientRequest) -> None:
        """Apply the authentication to the request.

        :param request: aiohttp request where the authentication should be applied
        """
        raise NotImplementedError()


class AuthenticatedClientRequest(ClientRequest):
    """Implementation of the ClientRequest that handles different types of authentication (not only BasicAuth)."""

    def update_auth(
        self, auth: Optional[Authentication], trust_env: bool = False
    ) -> None:
        """Override the authentication in the request to allow non-basic auth methods."""
        if not auth or not isinstance(auth, Authentication):
            return super().update_auth(auth, trust_env)

        auth.apply(self)


class BearerAuthentication(Authentication):
    """Bearer authentication class that adds the Bearer token to the request."""

    def __init__(self, tokens: list[BearerTokenForHost]):
        """Create a new BearerAuthentication instance.

        :param tokens: list of (host url, token) pairs. The token will be added to the request if the host url matches
            (including the scheme).
        """
        self.tokens = tokens

    def apply(self, request: ClientRequest) -> None:
        """Apply the authentication to the request.

        :param request: aiohttp request where the authentication should be applied
        """
        for token in self.tokens:
            if (
                request.url.host == token.host_url.host
                and request.url.scheme == token.host_url.scheme
            ):
                request.headers[hdrs.AUTHORIZATION] = f"Bearer {token.token}"
                break



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base types for invenio REST responses."""

from __future__ import annotations

from typing import Any


class Model:
    """Base model, which allows getting extra fields via normal dot operator."""

    _extra_data: dict[str, Any]

    def __getattr__(self, item: str) -> Any:  # noqa: ANN401
        """Get extra fields from the model_extra attribute."""
        if "_extra_data" not in self.__dict__:
            self._extra_data: dict[str, Any] = {}

        if item in self._extra_data:
            return self._extra_data[item]
        dash_item = item.replace("_", "-")
        if dash_item in self._extra_data:
            return self._extra_data[dash_item]

        raise AttributeError(f"{self.__class__.__name__} has no attribute {item}")

    def __getitem__(self, item: str) -> Any:
        """Get extra fields from the model_extra attribute."""
        return getattr(self, item)



import json as _json
import logging
from collections.abc import Callable
from datetime import datetime
from functools import partial
from typing import (
    Any,
    Optional,
    Protocol,
    Self,
    TypeVar,
    cast,
    dataclass_transform,
    get_args,
    get_origin,
)

from attrs import Attribute, define, fields
from cattrs import Converter
from cattrs.dispatch import StructureHook, UnstructureHook
from cattrs.gen import (
    AttributeOverride,
    make_dict_structure_fn,
    make_dict_unstructure_fn,
    override,
)
from yarl import URL

from .errors import StructureError, UnstructureError


def _remove_nulls(data):
    if isinstance(data, dict):
        for key, value in list(data.items()):
            if value is None:
                del data[key]
            else:
                _remove_nulls(value)
    elif isinstance(data, list):
        for idx in range(len(data) - 1, -1, -1):
            if data[idx] is None:
                del data[idx]
        for item in data:
            _remove_nulls(item)


class NullRemovingConverter(Converter):
    def unstructure(
        self, obj: Any, unstructure_as: Any = None, keep_nulls=False
    ) -> Any:
        try:
            ret = super().unstructure(obj, unstructure_as)
            if not keep_nulls:
                _remove_nulls(ret)
            return ret
        except Exception as e:
            raise UnstructureError(str(e)) from e

    def structure(self, obj: Any, type_: type) -> Any:
        try:
            return super().structure(obj, type_)
        except Exception as e:
            raise StructureError(str(e)) from e

converter = NullRemovingConverter()


class TypeHookStructureWrapper(Protocol):
    """Wrapper for structure hooks."""

    def __call__(self, data: Any, type_: type, previous: StructureHook) -> Any:
        """Wrap the structure hook."""
        ...


class TypeHookUnstructureWrapper(Protocol):
    """Wrapper for unstructure hooks."""

    def __call__(self, data: Any, previous: UnstructureHook) -> Any:
        """Wrap the unstructure hook."""
        ...


def structure_extra_data_hook(data: Any, type_: type, previous: StructureHook) -> Any:
    """Structure hook for extra data that moves extra data to a separate _extra_data attribute."""
    ret = previous(data, type)
    keys: set[str] = set()

    fld: Attribute
    for fld in fields(type_):
        if fld.alias:
            keys.add(fld.alias)
        else:
            keys.add(fld.name)
    extra_keys = data.keys() - keys
    extra_data = {k: data[k] for k in extra_keys}
    ret._extra_data = extra_data
    return ret


def unstructure_extra_data_hook(data: Any, previous: UnstructureHook) -> Any:
    """Unstructure hook for extra data that merges extra data from _extra_data attribute."""
    if "_extra_data" not in data.__dict__:
        return previous(data)
    extra_data = data._extra_data
    data = previous(data)
    data.pop("_extra_data", None)
    data.update(extra_data)
    return data


class TypeHookBuilder:
    """Builder for type hooks."""

    def __init__(self, t: type, allow_extra_data: bool) -> None:
        """Initialize the builder."""
        self._type = t
        self._structure_omitted_fields: list[str] = []
        self._unstructure_omitted_fields: list[str] = []
        self._rename_fields: dict[str, str] = {}
        self._structure_wrappers: list[TypeHookStructureWrapper] = []
        self._unstructure_wrappers: list[TypeHookUnstructureWrapper] = []
        self._allow_extra_data = allow_extra_data

    def omit_from_structure(self, *fields: str) -> Self:
        """Omit fields from structuring (deserialization)."""
        self._structure_omitted_fields.extend(fields)
        return self

    def omit_from_unstructure(self, *fields: str) -> Self:
        """Omit fields from unstructuring (serialization)."""
        self._unstructure_omitted_fields.extend(fields)
        return self

    def omit(self, *fields: str) -> Self:
        """Omit fields from both structuring and unstructuring."""
        self.omit_from_structure(*fields)
        self.omit_from_unstructure(*fields)
        return self

    def rename_field(self, serialized_name: str, model_name: str) -> Self:
        """Rename a field."""
        self._rename_fields[serialized_name] = model_name
        return self

    def add_structure_wrapper(self, wrapper: TypeHookStructureWrapper) -> Self:
        """Add a custom function for structuring."""
        self._structure_wrappers.append(wrapper)
        return self

    def add_unstructure_wrapper(self, wrapper: TypeHookUnstructureWrapper) -> Self:
        """Add a custom function for unstructuring."""
        self._unstructure_wrappers.append(wrapper)
        return self

    def build_type_hook(self) -> None:
        """Build and register the type hook."""
        if self._allow_extra_data:
            self._structure_wrappers.append(structure_extra_data_hook)
            self._unstructure_wrappers.append(unstructure_extra_data_hook)

        structure_overrides: dict[str, AttributeOverride] = {}
        for serialized_name, model_name in self._rename_fields.items():
            structure_overrides[model_name] = override(rename=serialized_name)
        for field in self._structure_omitted_fields:
            structure_overrides[field] = override(omit=True)

        unstructure_overrides: dict[str, AttributeOverride] = {}
        for serialized_name, model_name in self._rename_fields.items():
            unstructure_overrides[model_name] = override(rename=serialized_name)
        for field in self._unstructure_omitted_fields:
            unstructure_overrides[field] = override(omit=True)

        st_hook: StructureHook = make_dict_structure_fn(
            self._type, converter, **structure_overrides
        )

        unst_hook: UnstructureHook = make_dict_unstructure_fn(
            self._type, converter, **unstructure_overrides
        )
        if self._structure_wrappers:
            structure_wrappers: list[StructureHook] = []
            for structure_wrapper in self._structure_wrappers:
                if structure_wrappers:
                    structure_wrappers.append(
                        partial(structure_wrapper, previous=structure_wrappers[-1])
                    )
                else:
                    structure_wrappers.append(
                        partial(structure_wrapper, previous=st_hook)
                    )
            st_hook = structure_wrappers[-1]

        if self._unstructure_wrappers:
            unstructure_wrappers: list[UnstructureHook] = []
            for unstructure_wrapper in self._unstructure_wrappers:
                if unstructure_wrappers:
                    unstructure_wrappers.append(
                        partial(unstructure_wrapper, previous=unstructure_wrappers[-1])
                    )
                else:
                    unstructure_wrappers.append(
                        partial(unstructure_wrapper, previous=unst_hook)
                    )
            unst_hook = unstructure_wrappers[-1]

        converter.register_structure_hook(self._type, st_hook)
        converter.register_unstructure_hook(self._type, unst_hook)


class SerializationExtension:
    """Base class for serialization extensions."""

    def apply(self, builder: TypeHookBuilder) -> None:
        """Apply the extension to the builder."""
        pass


@define
class Omit(SerializationExtension):
    """Omit a field from serialization."""

    field: str
    from_structure: bool = True
    from_unstructure: bool = True

    def apply(self, builder: TypeHookBuilder) -> None:
        """Apply the extension to the builder."""
        if self.from_structure:
            builder.omit_from_structure(self.field)
        if self.from_unstructure:
            builder.omit_from_unstructure(self.field)


@define
class Rename(SerializationExtension):
    """Rename a field in serialization."""

    serialized_name: str
    model_name: str

    def apply(self, builder: TypeHookBuilder) -> None:
        """Apply the extension to the builder."""
        builder.rename_field(self.serialized_name, self.model_name)


@define
class WrapStructure(SerializationExtension):
    """Wrap a structure hook."""

    wrapper: TypeHookStructureWrapper

    def apply(self, builder: TypeHookBuilder) -> None:
        """Apply the extension to the builder."""
        builder.add_structure_wrapper(self.wrapper)


@define
class WrapUnstructure(SerializationExtension):
    """Wrap an unstructure hook."""

    wrapper: TypeHookUnstructureWrapper

    def apply(self, builder: TypeHookBuilder) -> None:
        """Apply the extension to the builder."""
        builder.add_unstructure_wrapper(self.wrapper)


C = TypeVar("C", bound=type)


@dataclass_transform()
def extend_serialization(
    *modifications: SerializationExtension,
    allow_extra_data: bool = False,
) -> Callable[[C], C]:
    """Add serialization rules to an attrs class."""

    def wrapper(t: C) -> C:
        """Wrap the class and register type for that."""
        th = TypeHookBuilder(t, allow_extra_data)
        for modification in modifications:
            modification.apply(th)
        th.build_type_hook()
        return t

    return wrapper


def structure_url_hook(value: str, type: type) -> URL:
    """Cattrs converter for URL."""
    return URL(value)


def unstructure_url_hook(value: URL) -> str:
    """Cattrs converter for URL."""
    return str(value)


converter.register_structure_hook(URL, structure_url_hook)
converter.register_unstructure_hook(URL, unstructure_url_hook)


def structure_datetime_hook(value: str, type: type) -> datetime:
    """Cattrs converter for URL."""
    return datetime.fromisoformat(value)


def unstructure_datetime_hook(value: datetime) -> str:
    """Cattrs converter for URL."""
    return value.isoformat()


converter.register_structure_hook(datetime, structure_datetime_hook)
converter.register_unstructure_hook(datetime, unstructure_datetime_hook)


log = logging.getLogger("invenio_nrp.client.deserialize")


def deserialize_rest_response[T](
    connection: Any,
    json_payload: bytes,
    result_class: type[T],
    etag: Optional[str],
) -> T:
    try:
        if get_origin(result_class) is list:
            arg_type = get_args(result_class)[0]
            return cast(
                "result_class",
                [
                    converter.structure(
                        {
                            **x,
                        },
                        arg_type,
                    )
                    for x in _json.loads(json_payload)
                ],
            )
        ret = converter.structure(
            {
                **_json.loads(json_payload),
                # "context": result_context,
            },
            result_class,
        )
        if hasattr(ret, "_etag"):
            ret._etag = etag
        return ret
    except Exception as e:
        import traceback
        traceback.print_exc()
        log.error("Error validating %s with %s", json_payload, result_class)
        raise e



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Helper methods to create output file names from path templates."""

from pathlib import Path
from typing import Any

from nrp_cmd.cli.base import OutputFormat
from nrp_cmd.converter import converter


def create_output_file_name(
    output_name: Path,
    obj_id: str,
    obj: Any,
    output_format: OutputFormat | None,
    **kwargs: Any,  # noqa: ANN401
) -> Path:
    """Create an output file name from a template."""
    # output name can contain variables inside {}. If it does, we will replace them with the values
    # from the record
    # we need to make sure that the expanded output name does not contain illegal combinations,
    # such as /../ or /./ or leading '/' - we strip them if that is the case

    output_parts: tuple[str, ...] = output_name.parts if output_name else ()
    is_absolute = output_name.is_absolute()

    transformed_parts = [
        format_part(part, obj_id, obj, output_format or OutputFormat.JSON, **kwargs)
        for part in output_parts
    ]
    transformed_parts = [part.replace("/", "") for part in transformed_parts]
    transformed_parts = [part.replace("\\", "") for part in transformed_parts]
    transformed_parts = [part.replace("..", "") for part in transformed_parts]
    transformed_parts = [part for part in transformed_parts if part]

    if is_absolute:
        output_name = Path("/", *transformed_parts)
    else:
        output_name = Path(*transformed_parts).absolute()

    return output_name


def format_part(
    part: str,
    obj_id: str,
    obj: Any,
    output_format: OutputFormat,
    **kwargs: Any,  # noqa: ANN401
) -> str:
    """Format a part of the output name.

    The part can contain variables in the form of {variable_name}. If the part contains
    both '{' and '}', it is considered a template and it will be formatted with the
    variables from the record and the kwargs.
    """
    if "{" in part and "}" in part:
        options: dict[str, str] = {
            "id": obj_id,
            "ext": f".{output_format.value}" if output_format else "table",
            **(converter.unstructure(obj) if obj else {}),
            **kwargs,
        }
        return part.format(**options)
    return part



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for updating metadata of files."""

from functools import partial
from typing import Optional

from rich.console import Console

from nrp_cmd.async_client import limit_connections
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.files.table_formatters import format_files_table
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.cli.records.metadata import read_metadata
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_verbosity
@with_model
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("key", type=str, help="Key for the file")
@argument_with_help("metadata", type=str, required=False, help="Metadata for the file")
@async_command
async def update_file_metadata(
    *,
    config: Config,
    repository: Optional[str] = None,
    record_id: str,
    key: str,
    metadata: Optional[str] = None,
    model: Model,
    out: Output,
) -> None:
    """Update the metadata of a file in a record."""
    console = Console()

    with limit_connections(10):
        (
            record,
            _record_id,
            _repository_config,
            _record_client,
            repository_client,
        ) = await read_record(
            record_id,
            repository,
            config,
            False,
            model.model,
            model.published,
            model.draft,
        )

        metadata = metadata or "{}"
        metadata_json = read_metadata(metadata)
        assert isinstance(metadata_json, dict), "Metadata must be a dictionary."

        files_client = repository_client.files
        files = await files_client.list(record)
        file = next((f for f in files if f.key == key), None)

        if not file:
            raise ValueError(
                f"File with key {key} not found in record {record_id}: {", ".join([f.key for f in files])}"
            )

        file.metadata = metadata_json
        updated_file = await files_client.update(file)

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(format_files_table, record, verbosity=out.verbosity),  # type: ignore # mypy can not infer the corredct type of the partial function
    ) as printer:
        printer.output([updated_file])



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line client for updating records."""

from collections.abc import Callable
from functools import partial
from typing import Any, Optional, Self, cast

import rich_click as click
from deepmerge import always_merger
from rich.console import Console

from nrp_cmd.async_client import (
    AsyncRecordsClient,
    get_async_client,
    get_repository_from_record_id,
)
from nrp_cmd.async_client.connection import AsyncConnection
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.records.metadata import read_metadata
from nrp_cmd.cli.records.table_formatters import format_record_table
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_verbosity
@with_model
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("metadata", type=str, help="Metadata")
@click.option("--replace/--merge", default=True, help="Replace or merge the metadata")
@click.option("--path", "-p", type=str, help="Path within the metadata")
@click.option("--model", type=str, help="Model name")
@click.option(
    "--published/--no-published", default=False, help="Include only published records"
)
@click.option("--draft/--no-draft", default=False, help="Include only drafts")
@async_command
async def update_record(
    *,
    config: Config,
    repository: Optional[str],
    record_id: str,
    metadata: str,
    replace: bool = True,
    path: Optional[str] = None,
    model: Model,
    out: Output,
) -> None:
    """Update a record with new metadata."""
    console = Console()

    if record_id.startswith("@"):
        record_id = record_id[1:]
        loaded_record_ids = config.load_variables()[record_id]
        if len(loaded_record_ids) != 1:
            raise ValueError(
                f"Variable {record_id} is not a single record ID: {loaded_record_ids}"
            )
        record_id = loaded_record_ids[0]

    metadata_json = read_metadata(metadata)
    record_url, repository_config = await get_repository_from_record_id(
        AsyncConnection(), record_id, config, repository
    )

    published = model.published
    draft = model.draft

    client = await get_async_client(repository, config=config)

    records_api: AsyncRecordsClient = client.records.draft_records
    if model.model is not None:
        # normally we use draft records, as most of the use cases are for draft records
        records_api = records_api.with_model(model.model)
    if model.model and not model.published and not model.draft:
        # make sure we have models information
        await client.get_repository_info()
        assert repository_config.info, "Repository info is missing"
        assert model.model, "Need to specify a model"
        repository_model = repository_config.info.models[model.model]
        if "drafts" in repository_model.features:
            draft = True
        else:
            published = True

    if published:
        records_api = client.records.published_records
    elif draft:
        records_api = client.records.draft_records

    record = await records_api.read(record_url)
    merge_metadata_at_path(
        record.metadata,
        metadata_json,
        replace,
        path,
    )

    record = await records_api.update(record)

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(format_record_table, verbosity=out.verbosity),  # type: ignore # mypy does not understand this
    ) as printer:
        printer.output(record)


def merge_metadata_at_path(
    metadata: Any,  # noqa: ANN401
    new_metadata: Any,  # noqa: ANN401
    replace: bool,
    path: str | None,
) -> dict[str, Any]:
    """Merge metadata at a path in a nested dictionary/list.

    :param metadata:         the whole metadata into which the new metadata should be merged
    :param new_metadata:     the new metadata to merge at path `path`
    :param replace:          whether to replace the old metadata at the path or merge them
    :param path:             the path to the metadata to merge
    :return:                 modified metadata
    """
    setters = InPathMDSetter.from_path(metadata, path or "")

    old = setters[-1].value
    if replace:
        if isinstance(old, dict):
            old.clear()
            cast(dict[str, Any], old).update(new_metadata)
        elif isinstance(old, list):
            old.clear()
            cast(list[Any], old).extend(new_metadata)
        else:
            old = new_metadata
    else:
        if isinstance(old, dict):
            always_merger.merge(cast(dict[str, Any], old), new_metadata)
        elif isinstance(old, list):
            cast(list[Any], old).extend(new_metadata)
        else:
            old = new_metadata
    setters[-1].value = old
    return setters[0].value


class InPathMDSetter:
    """A class for setting metadata at a path in a nested dictionary/list."""

    def __init__(
        self,
        metadata: dict[str, Any] | list[Any],
        parent: Self | None = None,
        parent_key: str | int | None = None,
    ):
        """Create a new InPathMDSetter object."""
        self.metadata: dict[str, Any] | list[Any] = metadata
        self.parent = parent
        self.parent_key = parent_key

    @classmethod
    def from_path(cls, metadata: dict[str, Any] | list[Any], path: str) -> list[Self]:
        """Create a list of InPathMDSetter objects representing the path to the metadata.

        Each object in the list represents a key in the path to the metadata with a getter
        and setter for the value at that key.
        """
        path_parts = path.split(".") if path else []
        path_parts = [x for x in path_parts if x]
        ret = [cls(metadata)]
        empty_factory: Callable[[], Any]

        def _empty_factory() -> None:
            return None

        for key_idx, key in enumerate(path_parts):

            empty_factory = _empty_factory

            if key_idx < len(path_parts) - 1:
                next_key = path_parts[key_idx + 1]
                empty_factory = list if next_key.isdigit() else dict
            ret.append(ret[-1]._get_key(key, empty_factory))
        return ret

    def _get_key(self, key: str, empty_factory: Callable[[], Any]) -> Self:
        cls: type[Self] = type(self)
        if isinstance(self.metadata, dict):
            if key not in self.metadata:
                return cls(empty_factory(), self, key)
            return cls(self.metadata[key], self, key)
        else:
            if int(key) >= len(self.metadata):
                return cls(empty_factory(), self, key)
            else:
                return cls(self.metadata[int(key)], self, key)

    @property
    def value(self) -> Any:  # noqa: ANN401
        """Get the value of the metadata at the path represented by this object."""
        return self.metadata

    @value.setter
    def value(self, value: Any) -> None:  # noqa: ANN401
        """Set the value of the metadata at the path represented by this object."""
        self.metadata = value
        if self.parent:
            assert self.parent_key is not None
            self.parent._set_key_in_parent(self.parent_key, self.metadata)

    def _set_key_in_parent(self, key: str | int, value: Any) -> None:  # noqa: ANN401
        assert self.parent_key is not None
        if isinstance(self.metadata, dict):
            assert isinstance(key, str)
            self.metadata[key] = value
        else:
            if int(key) < len(self.metadata):
                self.metadata[int(key)] = value
            else:
                self.metadata.append(value)
        if self.parent:
            self.parent._set_key_in_parent(self.parent_key, self.metadata)



#
# This file was generated from the asynchronous client at invenio/requests.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


from typing import Any, cast, overload, override

from yarl import URL

from ...types.info import RepositoryInfo
from ...types.records import Record
from ...types.requests import (
    Request,
    RequestList,
    RequestStatus,
    RequestType,
    RequestTypeList,
)
from ..base_client import SyncRequestsClient
from ..connection import SyncConnection


class SyncInvenioRequestsClient(SyncRequestsClient):
    """Invenio requests client implementation."""

    def __init__(self, connection: SyncConnection, info: RepositoryInfo):
        """Initialize the client."""
        self._connection = connection
        self._info = info

    @override
    def applicable_requests(
        self, topic: Record | URL, params: dict[str, str] | None = None
    ) -> RequestTypeList:
        """Return all requests that can be created on a given topic at this moment."""
        topic_url: URL
        if isinstance(topic, Record):
            if not topic.links.files:
                raise ValueError(
                    "The record does not have a files link, probably files are not enabled on it"
                )
            topic_url = topic.links.applicable_requests
        else:
            topic_url = topic / "applicable-requests"

        return self._connection.get(
            url=topic_url,
            result_class=RequestTypeList,
            params=params or {},
        )

    @overload
    def create(
        self,
        type_: RequestType,
        payload: dict[str, Any],
        submit: bool = False,
    ) -> Request: ...

    @overload
    def create(
        self,
        topic: Record | URL,
        type_: str,
        payload: dict[str, Any],
        submit: bool = False,
    ) -> Request: ...

    def create(  # type: ignore
        self,
        *args: Record | URL | RequestType | str | dict[str, Any],
        submit: bool = False,
    ) -> Request:
        """Create a new request of this type."""
        ...

        if len(args) == 2:
            type_ = cast(RequestType, args[0])
            payload = cast(dict[str, Any], args[1])
            assert isinstance(
                type_, RequestType
            ), "Invalid argument - expecting a request type"
        elif len(args) == 3:
            topic = cast(Record | URL, args[0])
            type_name = cast(str, args[1])
            payload = cast(dict[str, Any], args[2])
            request_types = self.applicable_requests(topic)
            type_ = request_types[type_name]
        else:
            raise ValueError("Invalid arguments")

        if not type_.links.actions.create:
            raise ValueError(f"The request type {type_.id} does not support creation")
        request: Request = self._connection.post(
            url=type_.links.actions.create,
            json=payload,
            result_class=Request,
        )
        if submit:
            return self.submit(request)
        return request

    @override
    def all(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Search for all requests the user has access to.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        requests_url: URL

        if topic is not None:
            if isinstance(topic, Record):
                requests_url = topic.links.requests
            elif isinstance(topic, URL):
                requests_url = topic / "requests"
            else:
                raise ValueError("Invalid topic")
        else:
            if not self._info.links.requests:
                raise ValueError("The repository does not have a requests link")
            requests_url = self._info.links.requests

        return self._connection.get(
            url=requests_url,
            result_class=RequestList,
            params=params or {},
        )

    @override
    def created(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all requests, that are created but not yet submitted.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "created"}
        )

    @override
    def submitted(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all submitted requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "submitted"}
        )

    @override
    def accepted(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all accepted requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "accepted"}
        )

    @override
    def declined(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all declined requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "declined"}
        )

    @override
    def expired(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all expired requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "expired"}
        )

    @override
    def cancelled(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all cancelled requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        return self.all(
            topic=topic, params={**(params or {}), "status": "cancelled"}
        )

    @override
    def read_request(self, request_id: str) -> Request:
        """Read a single request by its id.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        if self._info.links.requests is None:
            raise ValueError("The repository does not have a requests link")

        return self._connection.get(
            url=self._info.links.requests / request_id,
            result_class=Request,
        )

    @override
    def submit(
        self, request: Request | URL, payload: dict | None = None
    ) -> Request:
        """Submit the request.

        The request will be either passed to receivers, or auto-approved
        depending on the current workflow
        """
        return self._push_request(
            request=request,
            required_request_status=RequestStatus.CREATED,
            action="submit",
            payload=payload,
        )

    @override
    def accept(
        self, request: Request | URL, payload: dict | None = None
    ) -> Request:
        """Accept the submitted request."""
        return self._push_request(
            request=request,
            required_request_status=RequestStatus.SUBMITTED,
            action="accept",
            payload=payload,
        )

    @override
    def decline(
        self, request: Request | URL, payload: dict | None = None
    ) -> Request:
        """Decline the submitted request."""
        return self._push_request(
            request=request,
            required_request_status=RequestStatus.SUBMITTED,
            action="decline",
            payload=payload,
        )

    @override
    def cancel(
        self, request: Request | URL, payload: dict | None = None
    ) -> Request:
        """Cancel the request."""
        return self._push_request(
            request=request,
            required_request_status=RequestStatus.CREATED,
            action="cancel",
            payload=payload,
        )

    def _push_request(
        self,
        *,
        request: Request | URL,
        required_request_status: str,
        action: str,
        payload: dict | None,
    ) -> Request:
        """Push the request to the server."""
        if isinstance(request, URL):
            request = self._connection.get(
                url=request,
                result_class=Request,
            )

        if request.status != required_request_status:
            raise ValueError(
                f"Can {action} only requests with status {required_request_status}, not {request.status}"
            )

        action_link = getattr(request.links.actions, action)

        if not action_link:
            raise ValueError(f"You have no permission to {action} this request")

        return self._connection.post(
            url=action_link, json=payload or {}, result_class=Request
        )




#
# This file was generated from the asynchronous client at invenio/transfer/__init__.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Different transfers to the invenio repository."""

from .base import Transfer
from .registry import transfer_registry

__all__ = ("Transfer", "transfer_registry")




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Local transfer."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ...streams.progress import ProgressSource
from . import Transfer

if TYPE_CHECKING:
    from yarl import URL

    from ....progress import ProgressBar
    from ...connection import AsyncConnection
    from ...streams import DataSource
    from ..files import File


class LocalTransfer(Transfer):
    """Local transfer.

    This transfer copies a local file to the repository.
    The file will be stored in repository's primary storage (thus local)
    and the upload will be handled solely through the repository.
    """

    async def upload(
        self,
        connection: AsyncConnection,
        initialized_upload: File,
        source: DataSource,
        progress_bar: ProgressBar,
    ) -> None:
        """Upload the file."""
        if not initialized_upload.links.content:
            raise ValueError("The upload does not provide the content link.")

        headers = {}

        if "md5" in source.supported_checksums():
            md5_checksum = await source.checksum("md5")
            headers["Content-MD5"] = md5_checksum

        await connection.put_stream(
            url=initialized_upload.links.content,
            source=ProgressSource(source, progress_bar),
            headers=headers,
        )

    async def prepare(
        self,
        connection: AsyncConnection,
        files_link: URL,
        transfer_payload: dict,
        source: DataSource,
    ) -> None:
        """Prepare the transfer."""
        pass

    async def get_commit_payload(self, initialized_upload: File) -> dict:
        """Get payload for finalization of the successful upload."""
        return {}
